import logging

import salt.netapi
import salt.utils.json

logger = logging.getLogger(__name__)


class SaltInfo:
    """
    Class to  handle processing and publishing of "real time" Salt upates.
    """

    def __init__(self, handler):
        """
        handler is expected to be the server side end of a websocket
        connection.
        """
        self.handler = handler

        # These represent a "real time" view into Salt's jobs.
        self.jobs = {}

        # This represents a "real time" view of minions connected to Salt.
        self.minions = {}

    def publish_minions(self):
        """
        Publishes minions as a list of dicts.
        """
        minions = []

        for minion, minion_info in self.minions.items():
            curr_minion = {}
            curr_minion.update(minion_info)
            curr_minion.update({"id": minion})
            minions.append(curr_minion)

        ret = {"minions": minions}
        self.handler.send(salt.utils.json.dumps(ret), False)

    def publish(self, key, data):
        """
        Publishes the data to the event stream.
        """
        publish_data = {key: data}
        self.handler.send(salt.utils.json.dumps(publish_data), False)

    def process_minion_update(self, event_data):
        """
        Associate grains data with a minion and publish minion update
        """
        tag = event_data["tag"]
        event_info = event_data["data"]

        _, _, _, _, mid = tag.split("/")

        if not self.minions.get(mid, None):
            self.minions[mid] = {}

        minion = self.minions[mid]

        minion.update({"grains": event_info["return"]})

        self.publish_minions()

    def process_ret_job_event(self, event_data):
        """
        Process a /ret event returned by Salt for a particular minion.
        These events contain the returned results from a particular execution.
        """
        tag = event_data["tag"]
        event_info = event_data["data"]

        _, _, jid, _, mid = tag.split("/")
        job = self.jobs.setdefault(jid, {})

        minion = job.setdefault("minions", {}).setdefault(mid, {})
        minion.update({"return": event_info["return"]})
        minion.update({"retcode": event_info["retcode"]})
        minion.update({"success": event_info["success"]})

        job_complete = all(
            [minion["success"] for mid, minion in job["minions"].items()]
        )

        if job_complete:
            job["state"] = "complete"

        self.publish("jobs", self.jobs)

    def process_new_job_event(self, event_data):
        """
        Creates a new job with properties from the event data
        like jid, function, args, timestamp.

        Also sets the initial state to started.

        Minions that are participating in this job are also noted.

        """
        job = None
        tag = event_data["tag"]
        event_info = event_data["data"]
        minions = {}
        for mid in event_info["minions"]:
            minions[mid] = {"success": False}

        job = {
            "jid": event_info["jid"],
            "start_time": event_info["_stamp"],
            "minions": minions,  # is a dictionary keyed by mids
            "fun": event_info["fun"],
            "tgt": event_info["tgt"],
            "tgt_type": event_info["tgt_type"],
            "state": "running",
        }
        self.jobs[event_info["jid"]] = job
        self.publish("jobs", self.jobs)

    def process_key_event(self, event_data):
        """
        Tag: salt/key
        Data:
        {'_stamp': '2014-05-20T22:45:04.345583',
         'act': 'delete',
         'id': 'compute.home',
         'result': True}
        """

        tag = event_data["tag"]
        event_info = event_data["data"]

        if event_info["act"] == "delete":
            self.minions.pop(event_info["id"], None)
        elif event_info["act"] == "accept":
            self.minions.setdefault(event_info["id"], {})

        self.publish_minions()

    def process_presence_events(self, event_data, token, opts):
        """
        Check if any minions have connected or dropped.
        Send a message to the client if they have.
        """
        tag = event_data["tag"]
        event_info = event_data["data"]

        minions_detected = event_info["present"]
        curr_minions = self.minions.keys()

        changed = False

        # check if any connections were dropped
        dropped_minions = set(curr_minions) - set(minions_detected)

        for minion in dropped_minions:
            changed = True
            self.minions.pop(minion, None)

        # check if any new connections were made
        new_minions = set(minions_detected) - set(curr_minions)

        tgt = ",".join(new_minions)

        if tgt:
            changed = True
            client = salt.netapi.NetapiClient(opts)
            client.run(
                {
                    "fun": "grains.items",
                    "tgt": tgt,
                    "expr_type": "list",
                    "mode": "client",
                    "client": "local",
                    "asynchronous": "local_async",
                    "token": token,
                }
            )

        if changed:
            self.publish_minions()

    def process(self, salt_data, token, opts):
        """
        Process events and publish data
        """
        parts = salt_data["tag"].split("/")
        if len(parts) < 2:
            return

        # TBD: Simplify these conditional expressions
        if parts[1] == "job":
            if parts[3] == "new":
                self.process_new_job_event(salt_data)
                if salt_data["data"]["fun"] == "grains.items":
                    self.minions = {}
            elif parts[3] == "ret":
                self.process_ret_job_event(salt_data)
                if salt_data["data"]["fun"] == "grains.items":
                    self.process_minion_update(salt_data)
        if parts[1] == "key":
            self.process_key_event(salt_data)
        if parts[1] == "presence":
            self.process_presence_events(salt_data, token, opts)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import logging
from time import time_ns

from typing_extensions import deprecated

from opentelemetry import trace
from opentelemetry._events import Event
from opentelemetry._events import EventLogger as APIEventLogger
from opentelemetry._events import EventLoggerProvider as APIEventLoggerProvider
from opentelemetry._logs import (
    LogRecord,
    NoOpLogger,
    SeverityNumber,
    get_logger_provider,
)
from opentelemetry.sdk._logs import Logger, LoggerProvider
from opentelemetry.util.types import _ExtendedAttributes

_logger = logging.getLogger(__name__)


@deprecated(
    "You should use `Logger` instead. "
    "Deprecated since version 1.39.0 and will be removed in a future release."
)
class EventLogger(APIEventLogger):
    def __init__(
        self,
        logger_provider: LoggerProvider,
        name: str,
        version: str | None = None,
        schema_url: str | None = None,
        attributes: _ExtendedAttributes | None = None,
    ):
        super().__init__(
            name=name,
            version=version,
            schema_url=schema_url,
            attributes=attributes,
        )
        self._logger: Logger = logger_provider.get_logger(
            name, version, schema_url, attributes
        )

    def emit(self, event: Event) -> None:
        if isinstance(self._logger, NoOpLogger):
            # Do nothing if SDK is disabled
            return
        span_context = trace.get_current_span().get_span_context()

        log_record = LogRecord(
            timestamp=event.timestamp or time_ns(),
            observed_timestamp=None,
            trace_id=event.trace_id or span_context.trace_id,
            span_id=event.span_id or span_context.span_id,
            trace_flags=event.trace_flags or span_context.trace_flags,
            severity_text=None,
            severity_number=event.severity_number or SeverityNumber.INFO,
            body=event.body,
            attributes=event.attributes,
        )
        self._logger.emit(log_record)


@deprecated(
    "You should use `LoggerProvider` instead. "
    "Deprecated since version 1.39.0 and will be removed in a future release."
)
class EventLoggerProvider(APIEventLoggerProvider):
    def __init__(self, logger_provider: LoggerProvider | None = None):
        self._logger_provider = logger_provider or get_logger_provider()

    def get_event_logger(
        self,
        name: str,
        version: str | None = None,
        schema_url: str | None = None,
        attributes: _ExtendedAttributes | None = None,
    ) -> EventLogger:
        if not name:
            _logger.warning("EventLogger created with invalid name: %s", name)
        return EventLogger(
            self._logger_provider, name, version, schema_url, attributes
        )

    def shutdown(self):
        self._logger_provider.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        self._logger_provider.force_flush(timeout_millis)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "fixxpath",
    "findtext",
    "findattr",
    "findall",
    "findall_ignore_namespace",
    "findtext_ignore_namespace",
]


def fixxpath(xpath, namespace=None):
    # ElementTree wants namespaces in its xpaths, so here we add them.
    if not namespace:
        return xpath
    return "/".join(["{{{}}}{}".format(namespace, e) for e in xpath.split("/")])


def findtext(element, xpath, namespace=None, no_text_value=""):
    """
    :param no_text_value: Value to return if the provided element has no text
                          value.
    :type no_text_value: ``object``
    """
    value = element.findtext(fixxpath(xpath=xpath, namespace=namespace))

    if value == "":
        return no_text_value
    return value


def findtext_ignore_namespace(element, xpath, namespace=None, no_text_value=""):
    """
    Special version of findtext() which first tries to find the provided value using the provided
    namespace and in case no results are found we fallback to the xpath lookup without namespace.

    This is needed because some providers return some responses with namespace and some without.
    """

    result = findtext(
        element=element, xpath=xpath, namespace=namespace, no_text_value=no_text_value
    )

    if not result and namespace:
        result = findtext(element=element, xpath=xpath, namespace=None, no_text_value=no_text_value)

    return result


def findattr(element, xpath, namespace=None):
    return element.findtext(fixxpath(xpath=xpath, namespace=namespace))


def findall(element, xpath, namespace=None):
    return element.findall(fixxpath(xpath=xpath, namespace=namespace))


def findall_ignore_namespace(element, xpath, namespace=None):
    """
    Special version of findall() which first tries to find the provided value using the provided
    namespace and in case no results are found we fallback to the xpath lookup without namespace.

    This is needed because some providers return some responses with namespace and some without.
    """
    result = findall(element=element, xpath=xpath, namespace=namespace)

    if not result and namespace:
        result = findall(element=element, xpath=xpath, namespace=None)

    return result



import threading
from typing import Any

__all__ = ('ThreadPool', 'WorkerThread')

class TrueyZero:
    def __add__(self, other): ...
    def __radd__(self, other): ...

trueyzero: TrueyZero

class WorkerThread(threading.Thread):
    conn: Any
    server: Any
    ready: bool
    requests_seen: int
    bytes_read: int
    bytes_written: int
    start_time: Any
    work_time: int
    stats: Any
    def __init__(self, server): ...
    def run(self) -> None: ...

class ThreadPool:
    server: Any
    min: Any
    max: Any
    get: Any
    def __init__(
        self,
        server,
        min: int = ...,
        max: int = ...,
        accepted_queue_size: int = ...,
        accepted_queue_timeout: int = ...,
    ) -> None: ...
    def start(self) -> None: ...
    @property
    def idle(self): ...
    def put(self, obj) -> None: ...
    def grow(self, amount) -> None: ...
    def shrink(self, amount) -> None: ...
    def stop(self, timeout: int = ...) -> None: ...
    @property
    def qsize(self) -> int: ...



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


from opentelemetry.context import Context
from opentelemetry.util.types import Attributes


class Observation:
    """A measurement observed in an asynchronous instrument

    Return/yield instances of this class from asynchronous instrument callbacks.

    Args:
        value: The float or int measured value
        attributes: The measurement's attributes
        context: The measurement's context
    """

    def __init__(
        self,
        value: int | float,
        attributes: Attributes = None,
        context: Context | None = None,
    ) -> None:
        self._value = value
        self._attributes = attributes
        self._context = context

    @property
    def value(self) -> float | int:
        return self._value

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def context(self) -> Context | None:
        return self._context

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Observation)
            and self.value == other.value
            and self.attributes == other.attributes
            and self.context == other.context
        )

    def __repr__(self) -> str:
        return f"Observation(value={self.value}, attributes={self.attributes}, context={self.context})"



"""
Module to work with salt formula defaults files

"""

import copy
import logging
import os

import salt.fileclient
import salt.utils.data
import salt.utils.dictupdate as dictupdate
import salt.utils.files
import salt.utils.json
import salt.utils.url
import salt.utils.yaml

__virtualname__ = "defaults"

log = logging.getLogger(__name__)


def _mk_client():
    """
    Create a file client and add it to the context
    """
    return salt.fileclient.get_file_client(__opts__)


def _load(formula):
    """
    Generates a list of salt://<formula>/defaults.(json|yaml) files
    and fetches them from the Salt master.

    Returns first defaults file as python dict.
    """

    # Compute possibilities
    paths = []
    for ext in ("yaml", "json"):
        source_url = salt.utils.url.create(formula + "/defaults." + ext)
        paths.append(source_url)
    # Fetch files from master
    with _mk_client() as client:
        defaults_files = client.cache_files(paths)

    for file_ in defaults_files:
        if not file_:
            # Skip empty string returned by cp.fileclient.cache_files.
            continue

        suffix = file_.rsplit(".", 1)[-1]
        if suffix == "yaml":
            loader = salt.utils.yaml.safe_load
        elif suffix == "json":
            loader = salt.utils.json.load
        else:
            log.debug("Failed to determine loader for %r", file_)
            continue

        if os.path.exists(file_):
            log.debug("Reading defaults from %r", file_)
            with salt.utils.files.fopen(file_) as fhr:
                defaults = loader(fhr)
                log.debug("Read defaults %r", defaults)

            return defaults or {}


def get(key, default=""):
    """
    defaults.get is used much like pillar.get except that it will read
    a default value for a pillar from defaults.json or defaults.yaml
    files that are stored in the root of a salt formula.

    CLI Example:

    .. code-block:: bash

        salt '*' defaults.get core:users:root

    The defaults is computed from pillar key. The first entry is considered as
    the formula namespace.

    For example, querying ``core:users:root`` will try to load
    ``salt://core/defaults.yaml`` and ``salt://core/defaults.json``.
    """

    # Determine formula namespace from query
    if ":" in key:
        namespace, key = key.split(":", 1)
    else:
        namespace, key = key, None

    # Fetch and load defaults formula files from states.
    defaults = _load(namespace)

    # Fetch value
    if key:
        return salt.utils.data.traverse_dict_and_list(defaults, key, default)
    else:
        return defaults


def merge(dest, src, merge_lists=False, in_place=True, convert_none=True):
    """
    defaults.merge
        Allows deep merging of dicts in formulas.

    merge_lists : False
        If True, it will also merge lists instead of replace their items.

    in_place : True
        If True, it will merge into dest dict,
        if not it will make a new copy from that dict and return it.

    convert_none : True
        If True, it will convert src and dest to empty dicts if they are None.
        If True and dest is None but in_place is True, raises TypeError.
        If False it will make a new copy from that dict and return it.

        .. versionadded:: 3005

    CLI Example:

    .. code-block:: bash

        salt '*' defaults.merge '{a: b}' '{d: e}'

    It is more typical to use this in a templating language in formulas,
    instead of directly on the command-line.
    """
    # Force empty dicts if applicable (useful for cleaner templating)
    src = {} if (src is None and convert_none) else src
    if dest is None and convert_none:
        if in_place:
            raise TypeError("Can't perform in-place merge into NoneType")
        else:
            dest = {}

    if in_place:
        merged = dest
    else:
        merged = copy.deepcopy(dest)
    return dictupdate.update(merged, src, merge_lists=merge_lists)


def deepcopy(source):
    """
    defaults.deepcopy
        Allows deep copy of objects in formulas.

        By default, Python does not copy objects,
        it creates bindings between a target and an object.

    It is more typical to use this in a templating language in formulas,
    instead of directly on the command-line.
    """
    return copy.deepcopy(source)


def update(dest, defaults, merge_lists=True, in_place=True, convert_none=True):
    """
    defaults.update
        Allows setting defaults for group of data set e.g. group for nodes.

        This function is a combination of defaults.merge
        and defaults.deepcopy to avoid redundant in jinja.

        Example:

        .. code-block:: yaml

            group01:
              defaults:
                enabled: True
                extra:
                  - test
                  - stage
              nodes:
                host01:
                  index: foo
                  upstream: bar
                host02:
                  index: foo2
                  upstream: bar2

        .. code-block:: jinja

            {% do salt['defaults.update'](group01.nodes, group01.defaults) %}

        Each node will look like the following:

        .. code-block:: yaml

            host01:
              enabled: True
              index: foo
              upstream: bar
              extra:
                - test
                - stage

    merge_lists : True
        If True, it will also merge lists instead of replace their items.

    in_place : True
        If True, it will merge into dest dict.
        if not it will make a new copy from that dict and return it.

    convert_none : True
        If True, it will convert src and dest to empty dicts if they are None.
        If True and dest is None but in_place is True, raises TypeError.
        If False it will make a new copy from that dict and return it.

        .. versionadded:: 3005

    It is more typical to use this in a templating language in formulas,
    instead of directly on the command-line.
    """
    #  Force empty dicts if applicable here
    if in_place:
        if dest is None:
            raise TypeError("Can't perform in-place update into NoneType")
        else:
            nodes = dest
    else:
        dest = {} if (dest is None and convert_none) else dest
        nodes = deepcopy(dest)

    defaults = {} if (defaults is None and convert_none) else defaults

    for node_name, node_vars in nodes.items():
        defaults_vars = deepcopy(defaults)
        node_vars = merge(
            defaults_vars, node_vars, merge_lists=merge_lists, convert_none=convert_none
        )
        nodes[node_name] = node_vars

    return nodes



"""
Wheel command line tool (enables the ``python -m wheel`` syntax)
"""

from __future__ import annotations

import sys
from typing import NoReturn


def main() -> NoReturn:  # needed for console script
    if __spec__.parent == "":
        # To be able to run 'python wheel-0.9.whl/wheel':
        import os.path

        path = os.path.dirname(os.path.dirname(__file__))
        sys.path[0:0] = [path]

    from ._commands import main as cli_main

    sys.exit(cli_main())


if __name__ == "__main__":
    main()



"""
Retrieve Pillar data by doing a postgres query

.. versionadded:: 2017.7.0

:maturity: new
:depends: psycopg2
:platform: all

Complete Example
================

.. code-block:: yaml

    postgres:
      user: 'salt'
      pass: 'super_secret_password'
      db: 'salt_db'

    ext_pillar:
      - postgres:
          fromdb:
            query: 'SELECT col1,col2,col3,col4,col5,col6,col7
                      FROM some_random_table
                     WHERE minion_pattern LIKE %s'
            depth: 5
            as_list: True
            with_lists: [1,3]
"""

import logging
from contextlib import contextmanager

from salt.pillar.sql_base import SqlBaseExtPillar

# Set up logging
log = logging.getLogger(__name__)

try:
    import psycopg2

    HAS_POSTGRES = True
except ImportError:
    HAS_POSTGRES = False


def __virtual__():
    if not HAS_POSTGRES:
        return False
    return True


class POSTGRESExtPillar(SqlBaseExtPillar):
    """
    This class receives and processes the database rows from POSTGRES.
    """

    @classmethod
    def _db_name(cls):
        return "POSTGRES"

    def _get_options(self):
        """
        Returns options used for the POSTGRES connection.
        """
        defaults = {
            "host": "localhost",
            "user": "salt",
            "pass": "salt",
            "db": "salt",
            "port": 5432,
        }
        _options = {}
        _opts = __opts__.get("postgres", {})
        for attr in defaults:
            if attr not in _opts:
                log.debug("Using default for POSTGRES %s", attr)
                _options[attr] = defaults[attr]
                continue
            _options[attr] = _opts[attr]
        return _options

    @contextmanager
    def _get_cursor(self):
        """
        Yield a POSTGRES cursor
        """
        _options = self._get_options()
        conn = psycopg2.connect(
            host=_options["host"],
            user=_options["user"],
            password=_options["pass"],
            dbname=_options["db"],
            port=_options["port"],
        )
        cursor = conn.cursor()
        try:
            yield cursor
            log.debug("Connected to POSTGRES DB")
        except psycopg2.DatabaseError as err:
            log.exception("Error in ext_pillar POSTGRES: %s", err.args)
        finally:
            conn.close()

    def extract_queries(self, args, kwargs):  # pylint: disable=useless-super-delegation
        """
        This function normalizes the config block into a set of queries we
        can use.  The return is a list of consistently laid out dicts.
        """
        return super().extract_queries(args, kwargs)


def ext_pillar(minion_id, pillar, *args, **kwargs):
    """
    Execute queries against POSTGRES, merge and return as a dict
    """
    return POSTGRESExtPillar().fetch(minion_id, pillar, *args, **kwargs)



"""
Management of iptables
======================

This is an iptables-specific module designed to manage Linux firewalls. It is
expected that this state module, and other system-specific firewall states, may
at some point be deprecated in favor of a more generic ``firewall`` state.

.. code-block:: yaml

    httpd:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match:
            - state
            - comment
        - comment: "Allow HTTP"
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match:
            - state
            - comment
        - comment: "Allow HTTP"
        - connstate: NEW
        - source: '127.0.0.1'
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    .. Invert Rule
    httpd:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match:
            - state
            - comment
        - comment: "Allow HTTP"
        - connstate: NEW
        - source: '! 127.0.0.1'
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match:
            - state
            - comment
        - comment: "Allow HTTP"
        - connstate: NEW
        - source: 'not 127.0.0.1'
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.append:
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.append:
        - table: filter
        - family: ipv4
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dports:
            - 80
            - 443
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.insert:
        - position: 1
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.insert:
        - position: 1
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.delete:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.delete:
        - position: 1
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    httpd:
      iptables.delete:
        - table: filter
        - family: ipv6
        - chain: INPUT
        - jump: ACCEPT
        - match: state
        - connstate: NEW
        - dport: 80
        - protocol: tcp
        - sport: 1025:65535
        - save: True

    default to accept:
      iptables.set_policy:
        - chain: INPUT
        - policy: ACCEPT

.. note::

    Whereas iptables will accept ``-p``, ``--proto[c[o[l]]]`` as synonyms of
    ``--protocol``, if ``--proto`` appears in an iptables command after the
    appearance of ``-m policy``, it is interpreted as the ``--proto`` option of
    the policy extension (see the iptables-extensions(8) man page).


Example rules for IPSec policy:

.. code-block:: yaml

    accept_esp_in:
      iptables.append:
        - table: filter
        - chain: INPUT
        - jump: ACCEPT
        - source: 10.20.0.0/24
        - destination: 10.10.0.0/24
        - in-interface: eth0
        - match: policy
        - dir: in
        - pol: ipsec
        - reqid: 1
        - proto: esp
    accept_esp_forward_in:
      iptables.append:
        - use:
          - iptables: accept_esp_in
        - chain: FORWARD

    accept_esp_out:
      iptables.append:
        - table: filter
        - chain: OUTPUT
        - jump: ACCEPT
        - source: 10.10.0.0/24
        - destination: 10.20.0.0/24
        - out-interface: eth0
        - match: policy
        - dir: out
        - pol: ipsec
        - reqid: 1
        - proto: esp
    accept_esp_forward_out:
      iptables.append:
        - use:
          - iptables: accept_esp_out
        - chain: FORWARD

.. note::

    ``name`` is reserved for the Salt state name. To pass ``--name EXAMPLE`` to
    iptables, provide it with ``- name_: EXAMPLE``.

.. note::

    Various functions of the ``iptables`` module use the ``--check`` option. If
    the version of ``iptables`` on the target system does not include this
    option, an alternate version of this check will be performed using the
    output of iptables-save. This may have unintended consequences on legacy
    releases of ``iptables``.
"""

import copy

from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS


def __virtual__():
    """
    Only load if the locale module is available in __salt__
    """
    if "iptables.version" in __salt__:
        return True
    return (False, "iptables module could not be loaded")


def chain_present(name, table="filter", family="ipv4"):
    """
    .. versionadded:: 2014.1.0

    Verify the chain is exist.

    name
        A user-defined chain name.

    table
        The table to own the chain.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    chain_check = __salt__["iptables.check_chain"](table, name, family)
    if chain_check is True:
        ret["result"] = True
        ret["comment"] = "iptables {} chain is already exist in {} table for {}".format(
            name, table, family
        )
        return ret

    if __opts__["test"]:
        ret["comment"] = "iptables {} chain in {} table needs to be set for {}".format(
            name, table, family
        )
        return ret
    command = __salt__["iptables.new_chain"](table, name, family)
    if command is True:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "iptables {} chain in {} table create success for {}".format(
            name, table, family
        )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to create {} chain in {} table: {} for {}".format(
            name, table, command.strip(), family
        )
        return ret


def chain_absent(name, table="filter", family="ipv4"):
    """
    .. versionadded:: 2014.1.0

    Verify the chain is absent.

    table
        The table to remove the chain from

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    chain_check = __salt__["iptables.check_chain"](table, name, family)
    if not chain_check:
        ret["result"] = True
        ret["comment"] = (
            "iptables {} chain is already absent in {} table for {}".format(
                name, table, family
            )
        )
        return ret
    if __opts__["test"]:
        ret["comment"] = "iptables {} chain in {} table needs to be removed {}".format(
            name, table, family
        )
        return ret
    flush_chain = __salt__["iptables.flush"](table, name, family)
    if not flush_chain:
        command = __salt__["iptables.delete_chain"](table, name, family)
        if command is True:
            ret["changes"] = {"locale": name}
            ret["result"] = True
            ret["comment"] = (
                "iptables {} chain in {} table delete success for {}".format(
                    name, table, family
                )
            )
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete {} chain in {} table: {} for {}".format(
                name, table, command.strip(), family
            )
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush {} chain in {} table: {} for {}".format(
            name, table, flush_chain.strip(), family
        )
    return ret


def append(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 0.17.0

    Add a rule to the end of the specified chain.
    If the rule is already present anywhere in the chain, its position is
    not changed.

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    table
        The table that owns the chain which should be modified

    family
        Network family, ipv4 or ipv6.

    save
        If set to a true value, the new iptables rules for the given family
        will be saved to a file.

        If the value is True, rules are saved to an OS-dependent file
        that will be loaded during system startup, resulting in the
        firewall rule remaining active across reboots if possible.

        Note that loading the iptables rules during system startup
        may require non-default packages to be installed.
        On Debian-derived systems, the iptables-persistent
        package is required.

        If the value is a string, it is taken to be a filename to which
        the rules will be saved. Arranging for the rules to be loaded
        during system startup must be done separately.

    All other arguments are passed in with the same name as the long option
    that would normally be used for iptables, with one exception: ``--state`` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).

    Jump options that doesn't take arguments should be passed in with an empty
    string.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if "rules" in kwargs:
        ret["changes"]["locale"] = []
        comments = []
        save = False
        for rule in kwargs["rules"]:
            if "rules" in rule:
                del rule["rules"]
            if "__agg__" in rule:
                del rule["__agg__"]
            if "save" in rule and rule["save"]:
                save = True
                if rule["save"] is not True:
                    save_file = rule["save"]
                else:
                    save_file = True
                rule["save"] = False
            _ret = append(**rule)
            if "locale" in _ret["changes"]:
                ret["changes"]["locale"].append(_ret["changes"]["locale"])
            comments.append(_ret["comment"])
            ret["result"] = _ret["result"]
        if save:
            if save_file is True:
                save_file = None
            __salt__["iptables.save"](filename=save_file, family=family)
        if not ret["changes"]["locale"]:
            del ret["changes"]["locale"]
        ret["comment"] = "\n".join(comments)
        return ret

    if "__agg__" in kwargs:
        del kwargs["__agg__"]
    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    kwargs["name"] = name
    kwargs["table"] = table
    rule = __salt__["iptables.build_rule"](family=family, **kwargs)
    command = __salt__["iptables.build_rule"](
        full="True", family=family, command="A", **kwargs
    )
    if __salt__["iptables.check"](table, kwargs["chain"], rule, family) is True:
        ret["result"] = True
        ret["comment"] = "iptables rule for {} already set ({}) for {}".format(
            name, command.strip(), family
        )
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            saved_rules = __salt__["iptables.get_saved_rules"](
                conf_file=filename, family=family
            )
            _rules = __salt__["iptables.get_rules"](family=family)
            __rules = []
            for table in _rules:
                for chain in _rules[table]:
                    __rules.append(_rules[table][chain].get("rules"))
            __saved_rules = []
            for table in saved_rules:
                for chain in saved_rules[table]:
                    __saved_rules.append(saved_rules[table][chain].get("rules"))
            # Only save if rules in memory are different than saved rules
            if __rules != __saved_rules:
                out = __salt__["iptables.save"](filename=filename, family=family)
                ret["comment"] += "\nSaved iptables rule {} for {}\n{}\n{}".format(
                    name, family, command.strip(), out
                )
        return ret
    if __opts__["test"]:
        ret["comment"] = "iptables rule for {} needs to be set ({}) for {}".format(
            name, command.strip(), family
        )
        return ret
    if __salt__["iptables.append"](table, kwargs["chain"], rule, family):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set iptables rule for {} to: {} for {}".format(
            name, command.strip(), family
        )
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            out = __salt__["iptables.save"](filename=filename, family=family)
            ret["comment"] = "Set and saved iptables rule {} for {}\n{}\n{}".format(
                name, family, command.strip(), out
            )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to set iptables rule for {}.\nAttempted rule was {} for {}".format(
                name, command.strip(), family
            )
        )
        return ret


def insert(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.1.0

    Insert a rule into a chain. If the rule is already present anywhere
    in the chain, its position is not changed.

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    table
        The table that owns the chain that should be modified

    family
        Networking family, either ipv4 or ipv6

    position
        The numerical representation of where the rule should be inserted into
        the chain. Note that ``-1`` is not a supported position value.

    save
        If set to a true value, the new iptables rules for the given family
        will be saved to a file. See the ``append`` state for more details.

    All other arguments are passed in with the same name as the long option
    that would normally be used for iptables, with one exception: ``--state`` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).

    Jump options that doesn't take arguments should be passed in with an empty
    string.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if "rules" in kwargs:
        ret["changes"]["locale"] = []
        comments = []
        save = False
        for rule in kwargs["rules"]:
            if "rules" in rule:
                del rule["rules"]
            if "__agg__" in rule:
                del rule["__agg__"]
            if "save" in rule and rule["save"]:
                save = True
                if rule["save"] is not True:
                    save_file = rule["save"]
                else:
                    save_file = True
                rule["save"] = False
            _ret = insert(**rule)
            if "locale" in _ret["changes"]:
                ret["changes"]["locale"].append(_ret["changes"]["locale"])
            comments.append(_ret["comment"])
            ret["result"] = _ret["result"]
        if save:
            if save_file is True:
                save_file = None
            __salt__["iptables.save"](filename=save_file, family=family)
        if not ret["changes"]["locale"]:
            del ret["changes"]["locale"]
        ret["comment"] = "\n".join(comments)
        return ret

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    kwargs["name"] = name
    kwargs["table"] = table
    rule = __salt__["iptables.build_rule"](family=family, **kwargs)
    command = __salt__["iptables.build_rule"](
        full=True, family=family, command="I", **kwargs
    )
    if __salt__["iptables.check"](table, kwargs["chain"], rule, family) is True:
        ret["result"] = True
        ret["comment"] = "iptables rule for {} already set for {} ({})".format(
            name, family, command.strip()
        )
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            saved_rules = __salt__["iptables.get_saved_rules"](
                conf_file=filename, family=family
            )
            _rules = __salt__["iptables.get_rules"](family=family)
            __rules = []
            for table in _rules:
                for chain in _rules[table]:
                    __rules.append(_rules[table][chain].get("rules"))
            __saved_rules = []
            for table in saved_rules:
                for chain in saved_rules[table]:
                    __saved_rules.append(saved_rules[table][chain].get("rules"))
            # Only save if rules in memory are different than saved rules
            if __rules != __saved_rules:
                out = __salt__["iptables.save"](filename=filename, family=family)
                ret["comment"] += "\nSaved iptables rule {} for {}\n{}\n{}".format(
                    name, family, command.strip(), out
                )
        return ret
    if __opts__["test"]:
        ret["comment"] = "iptables rule for {} needs to be set for {} ({})".format(
            name, family, command.strip()
        )
        return ret
    if not __salt__["iptables.insert"](
        table, kwargs["chain"], kwargs["position"], rule, family
    ):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set iptables rule for {} to: {} for {}".format(
            name, command.strip(), family
        )
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            out = __salt__["iptables.save"](filename=filename, family=family)
            ret["comment"] = "Set and saved iptables rule {} for {}\n{}\n{}".format(
                name, family, command.strip(), out
            )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to set iptables rule for {}.\nAttempted rule was {}".format(
                name, command.strip()
            )
        )
        return ret


def delete(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.1.0

    Delete a rule from a chain if present. If the rule is already absent,
    this is not an error and nothing is changed.

    name
        A user-defined name to call this rule by in another part of a state or
        formula. This should not be an actual rule.

    table
        The table that owns the chain that should be modified

    family
        Networking family, either ipv4 or ipv6

    save
        If set to a true value, the new iptables rules for the given family
        will be saved to a file. See the ``append`` state for more details.

    All other arguments are passed in with the same name as the long option
    that would normally be used for iptables, with one exception: ``--state`` is
    specified as `connstate` instead of `state` (not to be confused with
    `ctstate`).

    Jump options that doesn't take arguments should be passed in with an empty
    string.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if "rules" in kwargs:
        ret["changes"]["locale"] = []
        comments = []
        save = False
        for rule in kwargs["rules"]:
            if "rules" in rule:
                del rule["rules"]
            if "__agg__" in rule:
                del rule["__agg__"]
            if "save" in rule and rule["save"]:
                if rule["save"] is not True:
                    save_file = rule["save"]
                else:
                    save_file = True
                rule["save"] = False
            _ret = delete(**rule)
            if "locale" in _ret["changes"]:
                ret["changes"]["locale"].append(_ret["changes"]["locale"])
            comments.append(_ret["comment"])
            ret["result"] = _ret["result"]
        if save:
            if save_file is True:
                save_file = None
            __salt__["iptables.save"](filename=save_file, family=family)
        if not ret["changes"]["locale"]:
            del ret["changes"]["locale"]
        ret["comment"] = "\n".join(comments)
        return ret

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]
    kwargs["name"] = name
    kwargs["table"] = table
    rule = __salt__["iptables.build_rule"](family=family, **kwargs)
    command = __salt__["iptables.build_rule"](
        full=True, family=family, command="D", **kwargs
    )

    if not __salt__["iptables.check"](table, kwargs["chain"], rule, family) is True:
        if "position" not in kwargs:
            ret["result"] = True
            ret["comment"] = "iptables rule for {} already absent for {} ({})".format(
                name, family, command.strip()
            )
            return ret
    if __opts__["test"]:
        ret["comment"] = "iptables rule for {} needs to be deleted for {} ({})".format(
            name, family, command.strip()
        )
        return ret

    if "position" in kwargs:
        result = __salt__["iptables.delete"](
            table, kwargs["chain"], family=family, position=kwargs["position"]
        )
    else:
        result = __salt__["iptables.delete"](
            table, kwargs["chain"], family=family, rule=rule
        )

    if not result:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = f"Delete iptables rule for {name} {command.strip()}"
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            out = __salt__["iptables.save"](filename=filename, family=family)
            ret["comment"] = "Deleted and saved iptables rule {} for {}\n{}\n{}".format(
                name, family, command.strip(), out
            )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = (
            "Failed to delete iptables rule for {}.\nAttempted rule was {}".format(
                name, command.strip()
            )
        )
        return ret


def set_policy(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.1.0

    Sets the default policy for iptables firewall tables

    table
        The table that owns the chain that should be modified

    family
        Networking family, either ipv4 or ipv6

    policy
        The requested table policy

    save
        If set to a true value, the new iptables rules for the given family
        will be saved to a file. See the ``append`` state for more details.

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]

    if (
        __salt__["iptables.get_policy"](table, kwargs["chain"], family)
        == kwargs["policy"]
    ):
        ret["result"] = True
        ret["comment"] = (
            "iptables default policy for chain {} on table {} for {} already set to {}".format(
                kwargs["chain"], table, family, kwargs["policy"]
            )
        )
        return ret
    if __opts__["test"]:
        ret["comment"] = (
            "iptables default policy for chain {} on table {} for {} needs to be set"
            " to {}".format(kwargs["chain"], table, family, kwargs["policy"])
        )
        return ret
    if not __salt__["iptables.set_policy"](
        table, kwargs["chain"], kwargs["policy"], family
    ):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Set default policy for {} to {} family {}".format(
            kwargs["chain"], kwargs["policy"], family
        )
        if "save" in kwargs and kwargs["save"]:
            if kwargs["save"] is not True:
                filename = kwargs["save"]
            else:
                filename = None
            __salt__["iptables.save"](filename=filename, family=family)
            ret["comment"] = (
                "Set and saved default policy for {} to {} family {}".format(
                    kwargs["chain"], kwargs["policy"], family
                )
            )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to set iptables default policy"
        return ret


def flush(name, table="filter", family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.1.0

    Flush current iptables state

    table
        The table that owns the chain that should be modified

    family
        Networking family, either ipv4 or ipv6

    chain
        The chain to be flushed. All the chains in the table if none is given.


    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in kwargs:
            del kwargs[ignore]

    if "chain" not in kwargs:
        kwargs["chain"] = ""
    if __opts__["test"]:
        ret["comment"] = (
            "iptables rules in {} table {} chain {} family needs to be flushed".format(
                name, table, family
            )
        )
        return ret
    if not __salt__["iptables.flush"](table, kwargs["chain"], family):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = "Flush iptables rules in {} table {} chain {} family".format(
            table, kwargs["chain"], family
        )
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush iptables rules"
        return ret


def mod_aggregate(low, chunks, running):
    """
    The mod_aggregate function which looks up all rules in the available
    low chunks and merges them into a single rules ref in the present low data
    """
    rules = []
    agg_enabled = [
        "append",
        "insert",
    ]
    if low.get("fun") not in agg_enabled:
        return low
    for chunk in chunks:
        tag = __utils__["state.gen_tag"](chunk)
        if tag in running:
            # Already ran the iptables state, skip aggregation
            continue
        if chunk.get("state") == "iptables":
            if "__agg__" in chunk:
                continue
            # Check for the same function
            if chunk.get("fun") != low.get("fun"):
                continue

            if chunk not in rules:
                rules.append(copy.deepcopy(chunk))
                chunk["__agg__"] = True

    if rules:
        if "rules" in low:
            low["rules"].extend(rules)
        else:
            low["rules"] = rules
    return low



import asyncio
import logging
import multiprocessing
import os
import socket
import time
import warnings

import aiohttp
import aiohttp.web
import tornado.ioloop

import salt.payload
import salt.transport.base
import salt.transport.frame
import salt.utils.asynchronous
from salt.transport.tcp import (
    USE_LOAD_BALANCER,
    LoadBalancerServer,
    _get_bind_addr,
    _get_socket,
    _set_tcp_keepalive,
)

log = logging.getLogger(__name__)


class PublishClient(salt.transport.base.PublishClient):
    """
    Tornado based TCP Pub Client
    """

    ttype = "ws"

    async_methods = [
        "connect",
        "connect_uri",
        "recv",
        # "close",
    ]
    close_methods = [
        "close",
    ]

    def __init__(self, opts, io_loop, **kwargs):  # pylint: disable=W0231
        self.opts = opts
        if io_loop is None:
            io_loop = tornado.ioloop.IOLoop.current()
        self.io_loop = io_loop
        self.asyncio_loop = salt.utils.asynchronous.aioloop(io_loop)

        self.connected = False
        self._closing = False
        self._closing = False
        self._closed = False

        self.backoff = opts.get("tcp_reconnect_backoff", 1)
        self.poller = None

        self.host = kwargs.get("host", None)
        self.port = kwargs.get("port", None)
        self.path = kwargs.get("path", None)
        self.ssl = kwargs.get("ssl", None)
        self.source_ip = self.opts.get("source_ip")
        self.source_port = self.opts.get("source_publish_port")
        self.connect_callback = None
        self.disconnect_callback = None
        if self.host is None and self.port is None:
            if self.path is None:
                raise Exception("A host and port or a path must be provided")
        elif self.host and self.port:
            if self.path:
                raise Exception("A host and port or a path must be provided, not both")
        self._ws = None
        self._session = None
        self._closing = False
        self.on_recv_task = None

    def close(self):
        if self._closing:
            return
        self._closing = True
        # Cancel the receive task but don't await it (like TCP does)
        if self.on_recv_task:
            self.on_recv_task.cancel()
            self.on_recv_task = None
        # Schedule async cleanup but don't wait for it
        if self._session is not None or self._ws is not None:
            self.asyncio_loop.create_task(self._async_cleanup())
        self._closed = True

    async def _async_cleanup(self):
        """Background cleanup of async resources"""
        try:
            if self._session is not None:
                await self._session.close()
                self._session = None
            if self._ws is not None:
                await self._ws.close()
                self._ws = None
        except Exception:  # pylint: disable=broad-except
            pass  # Cleanup is best-effort

    # pylint: disable=W1701
    def __del__(self):
        if not self._closing:
            warnings.warn(
                "unclosed publish client {self!r}", ResourceWarning, source=self
            )

    # pylint: enable=W1701
    def _decode_messages(self, messages):
        if not isinstance(messages, dict):
            body = salt.payload.loads(messages)
        else:
            body = messages
        return body

    async def getstream(self, **kwargs):
        if self.source_ip or self.source_port:
            kwargs.update(source_ip=self.source_ip, source_port=self.source_port)
        ws = None
        session = None
        start = time.monotonic()
        timeout = kwargs.get("timeout", None)
        while ws is None and (not self._closed and not self._closing):
            try:
                ctx = None
                if self.ssl is not None:
                    ctx = salt.transport.base.ssl_context(self.ssl, server_side=False)
                if self.host and self.port:
                    conn = aiohttp.TCPConnector()
                    session = aiohttp.ClientSession(connector=conn)
                    if self.ssl:
                        url = f"https://{self.host}:{self.port}/ws"
                    else:
                        url = f"http://{self.host}:{self.port}/ws"
                else:
                    conn = aiohttp.UnixConnector(path=self.path)
                    session = aiohttp.ClientSession(connector=conn)
                    if self.ssl:
                        url = "https://ipc.saltproject.io/ws"
                    else:
                        url = "http://ipc.saltproject.io/ws"
                log.debug("pub client connect %r %r", url, ctx)
                ws = await asyncio.wait_for(
                    session.ws_connect(url, ssl=ctx),
                    timeout if timeout is not None else 5,
                )
                # For SSL connections, give handshake time to complete and fail if invalid
                if ws and self.ssl:
                    await asyncio.sleep(0.1)
                    if ws.closed:
                        log.debug("WS closed after SSL handshake")
                        ws = None
                        await session.close()
                        continue
            except Exception as exc:  # pylint: disable=broad-except
                log.warning(
                    "WS Message Client encountered an exception while connecting to"
                    " %s:%s %s: %r, will reconnect in %d seconds",
                    self.host,
                    self.port,
                    self.path,
                    exc,
                    self.backoff,
                )
                if session:
                    await session.close()
                if timeout and time.monotonic() - start > timeout:
                    break
                await asyncio.sleep(self.backoff)
        return ws, session

    async def _connect(self, timeout=None):
        if self._ws is None:
            self._ws, self._session = await self.getstream(timeout=timeout)
            if self.connect_callback:
                self.connect_callback(True)  # pylint: disable=not-callable
            self.connected = True

    async def connect(
        self,
        port=None,
        connect_callback=None,
        disconnect_callback=None,
        timeout=None,
    ):
        if port is not None:
            self.port = port
        if connect_callback:
            self.connect_callback = None
        if disconnect_callback:
            self.disconnect_callback = None
        await self._connect(timeout=timeout)

    async def send(self, msg):
        await self.message_client.send(msg, reply=False)

    async def recv(self, timeout=None):
        while self._ws is None:
            await self.connect()
            await asyncio.sleep(0.001)
        if timeout == 0:
            try:
                raw_msg = await asyncio.wait_for(self._ws.receive(), 0.0001)
            except TimeoutError:
                return
            if raw_msg.type == aiohttp.WSMsgType.TEXT:
                if raw_msg.data == "close":
                    await self._ws.close()
            if raw_msg.type == aiohttp.WSMsgType.BINARY:
                return salt.payload.loads(raw_msg.data, raw=True)
            elif raw_msg.type == aiohttp.WSMsgType.ERROR:
                log.error(
                    "ws connection closed with exception %s", self._ws.exception()
                )
        elif timeout:
            return await asyncio.wait_for(self.recv(), timeout=timeout)
        else:
            while True:
                raw_msg = await self._ws.receive()
                if raw_msg.type == aiohttp.WSMsgType.TEXT:
                    if raw_msg.data == "close":
                        await self._ws.close()
                if raw_msg.type == aiohttp.WSMsgType.BINARY:
                    return salt.payload.loads(raw_msg.data, raw=True)
                elif raw_msg.type == aiohttp.WSMsgType.ERROR:
                    log.error(
                        "ws connection closed with exception %s",
                        self._ws.exception(),
                    )

    async def on_recv_handler(self, callback):
        while not self._ws:
            await asyncio.sleep(0.003)
        while True:
            msg = await self.recv()
            await callback(msg)

    def on_recv(self, callback):
        """
        Register a callback for received messages (that we didn't initiate)
        """
        if self.on_recv_task:
            # XXX: We are not awaiting this canceled task. This still needs to
            # be addressed.
            self.on_recv_task.cancel()
        if callback is None:
            self.on_recv_task = None
        else:
            self.on_recv_task = asyncio.create_task(self.on_recv_handler(callback))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class PublishServer(salt.transport.base.DaemonizedPublishServer):
    """ """

    # TODO: opts!
    # Based on default used in tornado.netutil.bind_sockets()
    backlog = 128
    async_methods = [
        "publish",
        # "close",
    ]
    close_methods = [
        "close",
    ]

    def __init__(
        self,
        opts,
        pub_host=None,
        pub_port=None,
        pub_path=None,
        pull_host=None,
        pull_port=None,
        pull_path=None,
        pull_path_perms=0o600,
        pub_path_perms=0o600,
        started=None,
        _shutdown=None,
        ssl=None,
    ):
        self.opts = opts
        self.pub_host = pub_host
        self.pub_port = pub_port
        self.pub_path = pub_path
        self.pull_host = pull_host
        self.pull_port = pull_port
        self.pull_path = pull_path
        self.pull_path_perms = pull_path_perms
        self.pub_path_perms = pub_path_perms
        self.ssl = ssl
        self.clients = set()
        self._run = None
        if _shutdown is None:
            self._shutdown = multiprocessing.Event()  # Cross-process shutdown signal
        else:
            self._shutdown = _shutdown
        self.pub_writer = None
        self.pub_reader = None
        self._connecting = None
        self.runner = None
        self.site = None
        self.puller = None
        if started is None:
            self.started = multiprocessing.Event()
        else:
            self.started = started

    @classmethod
    def support_ssl(cls):
        # Required from DaemonizedPublishServer
        return True

    def topic_support(self):
        # Required from DaemonizedPublishServer
        return not self.opts.get("order_masters", False)

    def __setstate__(self, state):
        self.__init__(**state)

    def __getstate__(self):
        return {
            "opts": self.opts,
            "pub_host": self.pub_host,
            "pub_port": self.pub_port,
            "pub_path": self.pub_path,
            "pull_host": self.pull_host,
            "pull_port": self.pull_port,
            "pull_path": self.pull_path,
            "pull_path_perms": self.pull_path_perms,
            "pub_path_perms": self.pub_path_perms,
            "ssl": self.ssl,
            "started": self.started,
            "_shutdown": self._shutdown,
        }

    def publish_daemon(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
        secrets=None,
        started=None,
    ):
        """
        Bind to the interface specified in the configuration file
        """
        if started is not None:
            self.started = started
        # Use asyncio event loop directly like ZeroMQ does
        io_loop = salt.utils.asynchronous.aioloop(tornado.ioloop.IOLoop())

        # Set up asyncio signal handler to stop the loop on SIGTERM
        import signal

        io_loop.add_signal_handler(signal.SIGTERM, io_loop.stop)

        publisher_task = io_loop.create_task(
            self.publisher(publish_payload, io_loop=io_loop)
        )
        try:
            io_loop.run_forever()
        finally:
            self.close()

    async def publisher(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
        io_loop=None,
    ):
        if io_loop is None:
            io_loop = tornado.ioloop.IOLoop.current()
        if self._run is None:
            self._run = asyncio.Event()

        # Monitor the multiprocessing shutdown event and stop the loop
        async def monitor_shutdown():
            loop = asyncio.get_running_loop()
            # Wait for shutdown signal in executor to avoid blocking
            await loop.run_in_executor(None, self._shutdown.wait)
            self._run.set()
            loop.stop()

        asyncio.create_task(monitor_shutdown())

        ctx = None
        if self.ssl is not None:
            ctx = salt.transport.base.ssl_context(self.ssl, server_side=True)
        if self.pub_path:
            server = aiohttp.web.Server(self.handle_request)
            self.runner = aiohttp.web.ServerRunner(server)
            await self.runner.setup()
            with salt.utils.files.set_umask(0o177):
                log.info("Publisher binding to socket %s", self.pub_path)
                self.site = aiohttp.web.UnixSite(
                    self.runner, self.pub_path, ssl_context=ctx
                )
                await self.site.start()
                os.chmod(self.pub_path, self.pub_path_perms)
        else:
            sock = _get_socket(self.opts)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            # _set_tcp_keepalive(sock, self.opts)
            sock.setblocking(0)
            sock.bind((self.pub_host, self.pub_port))
            sock.listen(self.backlog)
            server = aiohttp.web.Server(self.handle_request)
            self.runner = aiohttp.web.ServerRunner(server)
            await self.runner.setup()
            self.site = aiohttp.web.SockSite(self.runner, sock, ssl_context=ctx)
            log.info("Publisher binding to socket %s:%s", self.pub_host, self.pub_port)
            await self.site.start()

        self._pub_payload = publish_payload
        if self.pull_path:
            with salt.utils.files.set_umask(0o177):
                self.puller = await asyncio.start_unix_server(
                    self.pull_handler, self.pull_path
                )
                os.chmod(self.pull_path, self.pull_path_perms)
        else:
            self.puller = await asyncio.start_server(
                self.pull_handler, self.pull_host, self.pull_port
            )
        self.started.set()
        # Wait for shutdown signal instead of polling
        await self._run.wait()
        # Properly shut down aiohttp server
        await self.site.stop()
        await self.runner.cleanup()
        # Close the puller server
        self.puller.close()
        await self.puller.wait_closed()

    async def pull_handler(self, reader, writer):
        unpacker = salt.utils.msgpack.Unpacker(raw=True)
        while True:
            data = await reader.read(1024)
            unpacker.feed(data)
            for msg in unpacker:
                await self._pub_payload(msg)

    def pre_fork(self, process_manager, *args, **kwargs):
        """
        Do anything necessary pre-fork. Since this is on the master side this will
        primarily be used to create IPC channels and create our daemon process to
        do the actual publishing
        """
        process_manager.add_process(
            self.publish_daemon,
            args=[self.publish_payload],
            name=self.__class__.__name__,
        )

    async def handle_request(self, request):
        try:
            cert = request.get_extra_info("peercert")
        except AttributeError:
            pass
        else:
            if cert:
                name = salt.transport.base.common_name(cert)
                log.debug("Request client cert %r", name)
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        self.clients.add(ws)
        try:
            # Keep connection alive until client disconnects
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.ERROR:
                    log.error("ws connection closed with exception %s", ws.exception())
                    break
        finally:
            self.clients.discard(ws)
        return ws

    async def _connect(self):
        if self.pull_path:
            self.pub_reader, self.pub_writer = await asyncio.open_unix_connection(
                self.pull_path
            )
        else:
            self.pub_reader, self.pub_writer = await asyncio.open_connection(
                self.pull_host, self.pull_port
            )
        self._connecting = None

    def connect(self, timeout=None):
        log.debug("Connect pusher %s", self.pull_path)
        if self._connecting is None:
            self._connecting = asyncio.create_task(self._connect())
        return self._connecting

    async def publish(
        self, payload, **kwargs
    ):  # pylint: disable=invalid-overridden-method
        """
        Publish "load" to minions
        """
        if not self.pub_writer:
            await self.connect()
        self.pub_writer.write(salt.payload.dumps(payload, use_bin_type=True))
        await self.pub_writer.drain()

    async def publish_payload(self, payload, topic_list=None):
        payload = salt.payload.dumps(payload, use_bin_type=True)
        for ws in list(self.clients):
            try:
                await ws.send_bytes(payload)
            except ConnectionResetError:
                self.clients.discard(ws)

    def close(self):
        if self.pub_writer:
            self.pub_writer.close()
            self.pub_writer = None
            self.pub_reader = None
        # Signal shutdown across processes
        self._shutdown.set()
        if self._connecting:
            self._connecting.cancel()


class RequestServer(salt.transport.base.DaemonizedRequestServer):
    def __init__(self, opts):  # pylint: disable=W0231
        self.opts = opts
        self.site = None
        self.ssl = self.opts.get("ssl", None)
        self._run = None
        self._socket = None

    def pre_fork(self, process_manager, *args, **kwargs):
        """
        Pre-fork we need to create the zmq router device
        """
        if USE_LOAD_BALANCER:
            self.socket_queue = multiprocessing.Queue()
            process_manager.add_process(
                LoadBalancerServer,
                args=(self.opts, self.socket_queue),
                name="LoadBalancerServer",
            )
        elif not salt.utils.platform.is_windows():
            if self.opts.get("ipc_mode") == "ipc" and self.opts.get("workers_ipc_name"):
                self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                self._socket.setblocking(0)
                ipc_path = os.path.join(
                    self.opts["sock_dir"], self.opts["workers_ipc_name"]
                )
                if os.path.exists(ipc_path):
                    os.unlink(ipc_path)
                self._socket.bind(ipc_path)
                os.chmod(ipc_path, 0o600)
            else:
                self._socket = _get_socket(self.opts)
                self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                _set_tcp_keepalive(self._socket, self.opts)
                self._socket.setblocking(0)
                self._socket.bind(_get_bind_addr(self.opts, "ret_port"))

    def post_fork(self, message_handler, io_loop, **kwargs):
        """
        After forking we need to create all of the local sockets to listen to the
        router

        message_handler: function to call with your payloads
        """
        self.message_handler = message_handler
        self._run = asyncio.Event()
        self._started = asyncio.Event()

        # Convert to asyncio loop
        io_loop = salt.utils.asynchronous.aioloop(io_loop)

        async def server():
            server = aiohttp.web.Server(self.handle_message)
            runner = aiohttp.web.ServerRunner(server)
            await runner.setup()
            ctx = None
            if self.ssl is not None:
                ctx = salt.transport.base.ssl_context(self.ssl, server_side=True)
            self.site = aiohttp.web.SockSite(runner, self._socket, ssl_context=ctx)
            log.info("Worker binding to socket %s", self._socket)
            await self.site.start()
            self._started.set()
            # pause here for very long time by serving HTTP requests and
            # waiting for keyboard interruption
            # Wait for shutdown signal instead of polling
            await self._run.wait()
            await self.site.stop()
            self._socket.close()

        io_loop.create_task(server())

    async def handle_message(self, request):
        try:
            cert = request.get_extra_info("peercert")
        except AttributeError:
            pass
        else:
            if cert:
                name = salt.transport.base.common_name(cert)
                log.debug("Request client cert %r", name)
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                if msg.data == "close":
                    await ws.close()
            if msg.type == aiohttp.WSMsgType.BINARY:
                payload = salt.payload.loads(msg.data)
                reply = await self.message_handler(payload)
                await ws.send_bytes(salt.payload.dumps(reply))
            elif msg.type == aiohttp.WSMsgType.ERROR:
                log.error("ws connection closed with exception %s", ws.exception())
        return ws

    def close(self):
        if self._run is not None:
            self._run.set()  # Signal shutdown
        if self._socket is not None:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            self._socket = None

    async def forward_message(self, payload):
        """
        Forward a message into this transport's worker queue.

        Not implemented for WebSocket transport. Worker pool routing is only
        supported for ZeroMQ transport.
        """
        log.warning(
            "Worker pool message forwarding is not supported for WebSocket transport. "
            "Use ZeroMQ transport for worker pool routing."
        )
        return None


class RequestClient(salt.transport.base.RequestClient):

    ttype = "ws"

    def __init__(self, opts, io_loop):  # pylint: disable=W0231
        self.opts = opts
        self.sending = False
        self.ws = None
        self.session = None
        self.io_loop = salt.utils.asynchronous.aioloop(io_loop)
        self._closing = False
        self._closed = False
        self.ssl = self.opts.get("ssl", None)

    async def connect(self):  # pylint: disable=invalid-overridden-method
        ctx = None
        if self.ssl is not None:
            ctx = salt.transport.base.ssl_context(self.ssl, server_side=False)

        master_uri = self.opts.get("master_uri", "")
        if master_uri.startswith("ipc://"):
            socket_path = master_uri[6:]
            connector = aiohttp.UnixConnector(path=socket_path)
            self.session = aiohttp.ClientSession(connector=connector)
            URL = "http://localhost/ws"
        else:
            self.session = aiohttp.ClientSession()
            URL = self.get_master_uri(self.opts)

        log.debug("Connect to %s %s", URL, ctx)
        self.ws = await self.session.ws_connect(URL, ssl=ctx)

    async def send(self, load, timeout=60):
        if self.sending or self._closing:
            await asyncio.sleep(0.03)
        self.sending = True
        try:
            await self.connect()
            message = salt.payload.dumps(load)
            await self.ws.send_bytes(message)
            async for msg in self.ws:
                if msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR):
                    break
                data = salt.payload.loads(msg.data)
                break
            return data
        finally:
            self.sending = False

    async def _close(self):
        if self.ws is not None:
            await self.ws.close()
            self.ws = None
        if self.session is not None:
            await self.session.close()
            self.session = None
        self._closed = True

    def close(self):
        if self._closing:
            return
        self._closing = True
        self.close_task = asyncio.create_task(self._close())

    def get_master_uri(self, opts):
        if "master_uri" in opts:
            if self.opts.get("ssl", None):
                return opts["master_uri"].replace("tcp:", "https:", 1)
            return opts["master_uri"].replace("tcp:", "http:", 1)
        if self.opts.get("ssl", None):
            return f"https://{opts['master_ip']}:{opts['master_port']}/ws"
        return f"http://{opts['master_ip']}:{opts['master_port']}/ws"

    # pylint: disable=W1701
    def __del__(self):
        if not self._closing:
            warnings.warn(
                "Unclosed publish client {self!r}", ResourceWarning, source=self
            )

    # pylint: enable=W1701



"""
Management of firewalld

.. versionadded:: 2015.8.0

The following example applies changes to the public zone, blocks echo-reply
and echo-request packets, does not set the zone to be the default, enables
masquerading, and allows ports 22/tcp and 25/tcp.
It will be applied permanently and directly before restart/reload.

.. code-block:: yaml

    public:
      firewalld.present:
        - name: public
        - block_icmp:
          - echo-reply
          - echo-request
        - default: False
        - masquerade: True
        - ports:
          - 22/tcp
          - 25/tcp

The following example applies changes to the public zone, enables
masquerading and configures port forwarding TCP traffic from port 22
to 2222, and forwards TCP traffic from port 80 to 443 at 192.168.0.1.

.. code-block:: yaml

  my_zone:
    firewalld.present:
      - name: public
      - masquerade: True
      - port_fwd:
        - 22:2222:tcp
        - 80:443:tcp:192.168.0.1

The following example binds the public zone to interface eth0 and to all
packets coming from the 192.168.1.0/24 subnet. It also removes the zone
from all other interfaces or sources.

.. code-block:: yaml

  public:
    firewalld.present:
      - name: public
      - interfaces:
        - eth0
      - sources:
        - 192.168.1.0/24

Here, we define a new service that encompasses TCP ports 4505 4506:

.. code-block:: yaml

  saltmaster:
    firewalld.service:
      - name: saltmaster
      - ports:
        - 4505/tcp
        - 4506/tcp

To make this new service available in a zone, the following can be used, which
would allow access to the salt master from the 10.0.0.0/8 subnet:

.. code-block:: yaml

  saltzone:
    firewalld.present:
      - name: saltzone
      - services:
        - saltmaster
      - sources:
        - 10.0.0.0/8

Another way of implementing the same rule above using rich rules is demonstrated
here:

.. code-block:: yaml

  saltzone:
    firewalld.present:
      - name: saltzone
      - rich_rules:
        - rule service name="saltmaster" accept
      - sources:
        - 10.0.0.0/8

The format of rich rules is the same as:

.. code-block:: shell

  firewall-cmd --list-rich-rules

with an example output of:

.. code-block:: text

  rule protocol value="icmp" accept
  rule protocol value="ipv6-icmp" accept
  rule service name="snmp" accept
"""

import logging

import salt.utils.path
from salt.exceptions import CommandExecutionError
from salt.output import nested

log = logging.getLogger(__name__)


class ForwardingMapping:
    """
    Represents a port forwarding statement mapping a local port to a remote
    port for a specific protocol (TCP or UDP)
    """

    def __init__(self, srcport, destport, protocol, destaddr):
        self.srcport = srcport
        self.destport = destport
        self.protocol = protocol
        self.destaddr = destaddr

    def __eq__(self, other):
        return (
            self.srcport == other.srcport
            and self.destport == other.destport
            and self.protocol == other.protocol
            and self.destaddr == other.destaddr
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    # hash is needed for set operations
    def __hash__(self):
        return (
            hash(self.srcport)
            ^ hash(self.destport)
            ^ hash(self.protocol)
            ^ hash(self.destaddr)
        )

    def todict(self):
        """
        Returns a pretty dictionary meant for command line output.
        """
        return {
            "Source port": self.srcport,
            "Destination port": self.destport,
            "Protocol": self.protocol,
            "Destination address": self.destaddr,
        }


def _parse_forward(mapping):
    """
    Parses a port forwarding statement in the form used by this state:

    from_port:to_port:protocol[:destination]

    and returns a ForwardingMapping object
    """
    if len(mapping.split(":")) > 3:
        (srcport, destport, protocol, destaddr) = mapping.split(":")
    else:
        (srcport, destport, protocol) = mapping.split(":")
        destaddr = ""
    return ForwardingMapping(srcport, destport, protocol, destaddr)


def __virtual__():
    """
    Ensure the firewall-cmd is available
    """
    if salt.utils.path.which("firewall-cmd"):
        return True

    return (
        False,
        "firewall-cmd is not available, firewalld is probably not installed.",
    )


def present(
    name,
    block_icmp=None,
    prune_block_icmp=False,
    default=None,
    masquerade=None,
    ports=None,
    prune_ports=False,
    port_fwd=None,
    prune_port_fwd=False,
    services=None,
    prune_services=False,
    interfaces=None,
    prune_interfaces=False,
    sources=None,
    prune_sources=False,
    rich_rules=None,
    prune_rich_rules=False,
):
    """
    Ensure a zone has specific attributes.

    name
        The zone to modify.

    default : None
        Set this zone as the default zone if ``True``.

    masquerade : None
        Enable or disable masquerade for a zone. By default it will not change it.

    block_icmp : None
        List of ICMP types to block in the zone.

    prune_block_icmp : False
        If ``True``, remove all but the specified block_icmp from the zone.

    ports : None
        List of ports to add to the zone.

    prune_ports : False
        If ``True``, remove all but the specified ports from the zone.

    port_fwd : None
        List of port forwards to add to the zone.

    prune_port_fwd : False
        If ``True``, remove all but the specified port_fwd from the zone.

    services : None
        List of services to add to the zone.

    prune_services : False
        If ``True``, remove all but the specified services from the zone.
        .. note:: Currently defaults to True for compatibility, but will be changed to False in a future release.

    interfaces : None
        List of interfaces to add to the zone.

    prune_interfaces : False
        If ``True``, remove all but the specified interfaces from the zone.

    sources : None
        List of sources to add to the zone.

    prune_sources : False
        If ``True``, remove all but the specified sources from the zone.

    rich_rules : None
        List of rich rules to add to the zone.

    prune_rich_rules : False
        If ``True``, remove all but the specified rich rules from the zone.
    """
    ret = _present(
        name,
        block_icmp,
        prune_block_icmp,
        default,
        masquerade,
        ports,
        prune_ports,
        port_fwd,
        prune_port_fwd,
        services,
        prune_services,
        interfaces,
        prune_interfaces,
        sources,
        prune_sources,
        rich_rules,
        prune_rich_rules,
    )

    # Reload firewalld service on changes
    if ret["changes"] != {}:
        __salt__["firewalld.reload_rules"]()

    return ret


def service(name, ports=None, protocols=None):
    """
    Ensure the service exists and encompasses the specified ports and
    protocols.

    .. versionadded:: 2016.11.0
    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    if name not in __salt__["firewalld.get_services"]():
        __salt__["firewalld.new_service"](name, restart=False)

    ports = ports or []

    try:
        _current_ports = __salt__["firewalld.get_service_ports"](name)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    new_ports = set(ports) - set(_current_ports)
    old_ports = set(_current_ports) - set(ports)

    for port in new_ports:
        if not __opts__["test"]:
            try:
                __salt__["firewalld.add_service_port"](name, port)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

    for port in old_ports:
        if not __opts__["test"]:
            try:
                __salt__["firewalld.remove_service_port"](name, port)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

    if new_ports or old_ports:
        ret["changes"].update({"ports": {"old": _current_ports, "new": ports}})

    protocols = protocols or []

    try:
        _current_protocols = __salt__["firewalld.get_service_protocols"](name)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    new_protocols = set(protocols) - set(_current_protocols)
    old_protocols = set(_current_protocols) - set(protocols)

    for protocol in new_protocols:
        if not __opts__["test"]:
            try:
                __salt__["firewalld.add_service_protocol"](name, protocol)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

    for protocol in old_protocols:
        if not __opts__["test"]:
            try:
                __salt__["firewalld.remove_service_protocol"](name, protocol)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

    if new_protocols or old_protocols:
        ret["changes"].update(
            {"protocols": {"old": _current_protocols, "new": protocols}}
        )

    if ret["changes"] != {}:
        __salt__["firewalld.reload_rules"]()

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = f"'{name}' is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Configuration for '{name}' will change."
        return ret

    ret["comment"] = f"'{name}' was configured."
    return ret


def _normalize_rich_rules(rich_rules):
    """
    Make sure rich rules are normalized and attributes
    are quoted with double quotes so it matches the output
    from firewall-cmd

    Example:

    rule family="ipv4" source address="192.168.0.0/16" port port=22 protocol=tcp accept
    rule family="ipv4" source address="192.168.0.0/16" port port='22' protocol=tcp accept
    rule family='ipv4' source address='192.168.0.0/16' port port='22' protocol=tcp accept

    normalized to:

    rule family="ipv4" source address="192.168.0.0/16" port port="22" protocol="tcp" accept
    """
    normalized_rules = []
    for rich_rule in rich_rules:
        normalized_rule = ""
        for cmd in rich_rule.split(" "):
            cmd_components = cmd.split("=", 1)
            if len(cmd_components) == 2:
                assigned_component = cmd_components[1]
                if not assigned_component.startswith(
                    '"'
                ) and not assigned_component.endswith('"'):
                    if assigned_component.startswith(
                        "'"
                    ) and assigned_component.endswith("'"):
                        assigned_component = assigned_component[1:-1]
                    cmd_components[1] = f'"{assigned_component}"'
            normalized_rule = f"{normalized_rule} {'='.join(cmd_components)}"
        normalized_rules.append(normalized_rule.lstrip())
    return normalized_rules


def _present(
    name,
    block_icmp=None,
    prune_block_icmp=False,
    default=None,
    masquerade=None,
    ports=None,
    prune_ports=False,
    port_fwd=None,
    prune_port_fwd=False,
    services=None,
    # TODO: prune_services=False in future release
    # prune_services=False,
    prune_services=None,
    interfaces=None,
    prune_interfaces=False,
    sources=None,
    prune_sources=False,
    rich_rules=None,
    prune_rich_rules=False,
):
    """
    Ensure a zone has specific attributes.
    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    try:
        zones = __salt__["firewalld.get_zones"](permanent=True)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret

    if name not in zones:
        if not __opts__["test"]:
            try:
                __salt__["firewalld.new_zone"](name)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

        ret["changes"].update({name: {"old": zones, "new": name}})

    if block_icmp or prune_block_icmp:
        block_icmp = block_icmp or []
        new_icmp_types = []
        old_icmp_types = []

        try:
            _current_icmp_blocks = __salt__["firewalld.list_icmp_block"](
                name, permanent=True
            )
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        if block_icmp:
            try:
                _valid_icmp_types = __salt__["firewalld.get_icmp_types"](permanent=True)
            except CommandExecutionError as err:
                ret["comment"] = f"Error: {err}"
                return ret

            # log errors for invalid ICMP types in block_icmp input
            for icmp_type in set(block_icmp) - set(_valid_icmp_types):
                log.error("%s is an invalid ICMP type", icmp_type)
                block_icmp.remove(icmp_type)

            new_icmp_types = set(block_icmp) - set(_current_icmp_blocks)
            for icmp_type in new_icmp_types:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.block_icmp"](
                            name, icmp_type, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if prune_block_icmp:
            old_icmp_types = set(_current_icmp_blocks) - set(block_icmp)
            for icmp_type in old_icmp_types:
                # no need to check against _valid_icmp_types here, because all
                # elements in old_icmp_types are guaranteed to be in
                # _current_icmp_blocks, whose elements are inherently valid
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.allow_icmp"](
                            name, icmp_type, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_icmp_types or old_icmp_types:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_block_icmp:
                block_icmp = list(new_icmp_types | set(_current_icmp_blocks))
            ret["changes"].update(
                {"icmp_types": {"old": _current_icmp_blocks, "new": block_icmp}}
            )

    # that's the only parameter that can't be permanent or runtime, it's
    # directly both
    if default:
        try:
            default_zone = __salt__["firewalld.default_zone"]()
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret
        if name != default_zone:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.set_default_zone"](name)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret
            ret["changes"].update({"default": {"old": default_zone, "new": name}})

    try:
        masquerade_ret = __salt__["firewalld.get_masquerade"](name, permanent=True)
    except CommandExecutionError as err:
        ret["comment"] = f"Error: {err}"
        return ret
    if masquerade is not None:
        if masquerade and not masquerade_ret:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_masquerade"](name, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret
            ret["changes"].update(
                {
                    "masquerade": {
                        "old": masquerade_ret,
                        "new": "Masquerading successfully set.",
                    }
                }
            )
        elif not masquerade and masquerade_ret:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.remove_masquerade"](name, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret
            ret["changes"].update(
                {
                    "masquerade": {
                        "old": masquerade_ret,
                        "new": "Masquerading successfully disabled.",
                    }
                }
            )

    if ports or prune_ports:
        ports = ports or []
        try:
            _current_ports = __salt__["firewalld.list_ports"](name, permanent=True)
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        new_ports = set(ports) - set(_current_ports)
        old_ports = []

        for port in new_ports:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_port"](
                        name, port, permanent=True, force_masquerade=False
                    )
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_ports:
            old_ports = set(_current_ports) - set(ports)
            for port in old_ports:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_port"](name, port, permanent=True)
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_ports or old_ports:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_ports:
                ports = list(new_ports | set(_current_ports))
            ret["changes"].update({"ports": {"old": _current_ports, "new": ports}})

    if port_fwd or prune_port_fwd:
        port_fwd = port_fwd or []
        try:
            _current_port_fwd = __salt__["firewalld.list_port_fwd"](
                name, permanent=True
            )
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        port_fwd = [_parse_forward(fwd) for fwd in port_fwd]
        _current_port_fwd = [
            ForwardingMapping(
                srcport=fwd["Source port"],
                destport=fwd["Destination port"],
                protocol=fwd["Protocol"],
                destaddr=fwd["Destination address"],
            )
            for fwd in _current_port_fwd
        ]

        new_port_fwd = set(port_fwd) - set(_current_port_fwd)
        old_port_fwd = []

        for fwd in new_port_fwd:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_port_fwd"](
                        name,
                        fwd.srcport,
                        fwd.destport,
                        fwd.protocol,
                        fwd.destaddr,
                        permanent=True,
                        force_masquerade=False,
                    )
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_port_fwd:
            old_port_fwd = set(_current_port_fwd) - set(port_fwd)
            for fwd in old_port_fwd:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_port_fwd"](
                            name,
                            fwd.srcport,
                            fwd.destport,
                            fwd.protocol,
                            fwd.destaddr,
                            permanent=True,
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_port_fwd or old_port_fwd:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_port_fwd:
                port_fwd = list(new_port_fwd | set(_current_port_fwd))
            ret["changes"].update(
                {
                    "port_fwd": {
                        "old": [fwd.todict() for fwd in _current_port_fwd],
                        "new": [fwd.todict() for fwd in port_fwd],
                    }
                }
            )

    if services or prune_services:
        services = services or []
        try:
            _current_services = __salt__["firewalld.list_services"](
                name, permanent=True
            )
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        new_services = set(services) - set(_current_services)
        old_services = []

        for new_service in new_services:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_service"](new_service, name, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_services:
            old_services = set(_current_services) - set(services)
            for old_service in old_services:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_service"](
                            old_service, name, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_services or old_services:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_services:
                services = list(new_services | set(_current_services))
            ret["changes"].update(
                {"services": {"old": _current_services, "new": services}}
            )

    if interfaces or prune_interfaces:
        interfaces = interfaces or []
        try:
            _current_interfaces = __salt__["firewalld.get_interfaces"](
                name, permanent=True
            )
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        new_interfaces = set(interfaces) - set(_current_interfaces)
        old_interfaces = []

        for interface in new_interfaces:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_interface"](name, interface, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_interfaces:
            old_interfaces = set(_current_interfaces) - set(interfaces)
            for interface in old_interfaces:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_interface"](
                            name, interface, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_interfaces or old_interfaces:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_interfaces:
                interfaces = list(new_interfaces | set(_current_interfaces))
            ret["changes"].update(
                {"interfaces": {"old": _current_interfaces, "new": interfaces}}
            )

    if sources or prune_sources:
        sources = sources or []
        try:
            _current_sources = __salt__["firewalld.get_sources"](name, permanent=True)
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        new_sources = set(sources) - set(_current_sources)
        old_sources = []

        for source in new_sources:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_source"](name, source, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_sources:
            old_sources = set(_current_sources) - set(sources)
            for source in old_sources:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_source"](
                            name, source, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_sources or old_sources:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_sources:
                sources = list(new_sources | set(_current_sources))
            ret["changes"].update(
                {"sources": {"old": _current_sources, "new": sources}}
            )

    if rich_rules or prune_rich_rules:
        rich_rules = rich_rules or []
        rich_rules = _normalize_rich_rules(rich_rules)
        try:
            _current_rich_rules = __salt__["firewalld.get_rich_rules"](
                name, permanent=True
            )
        except CommandExecutionError as err:
            ret["comment"] = f"Error: {err}"
            return ret

        new_rich_rules = set(rich_rules) - set(_current_rich_rules)
        old_rich_rules = []

        for rich_rule in new_rich_rules:
            if not __opts__["test"]:
                try:
                    __salt__["firewalld.add_rich_rule"](name, rich_rule, permanent=True)
                except CommandExecutionError as err:
                    ret["comment"] = f"Error: {err}"
                    return ret

        if prune_rich_rules:
            old_rich_rules = set(_current_rich_rules) - set(rich_rules)
            for rich_rule in old_rich_rules:
                if not __opts__["test"]:
                    try:
                        __salt__["firewalld.remove_rich_rule"](
                            name, rich_rule, permanent=True
                        )
                    except CommandExecutionError as err:
                        ret["comment"] = f"Error: {err}"
                        return ret

        if new_rich_rules or old_rich_rules:
            # If we're not pruning, include current items in new output so it's clear
            # that they're still present
            if not prune_rich_rules:
                rich_rules = list(new_rich_rules | set(_current_rich_rules))
            ret["changes"].update(
                {"rich_rules": {"old": _current_rich_rules, "new": rich_rules}}
            )

    # No changes
    if ret["changes"] == {}:
        ret["result"] = True
        ret["comment"] = f"'{name}' is already in the desired state."
        return ret

    # test=True and changes predicted
    if __opts__["test"]:
        ret["result"] = None
        # build comment string
        nested.__opts__ = __opts__
        comment = []
        comment.append(f"Configuration for '{name}' will change:")
        comment.append(nested.output(ret["changes"]).rstrip())
        ret["comment"] = "\n".join(comment)
        ret["changes"] = {}
        return ret

    # Changes were made successfully
    ret["result"] = True
    ret["comment"] = f"'{name}' was configured."
    return ret



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

POOL_NAME: Final = "pool.name"
"""
Deprecated: Replaced by `db.client.connection.pool.name`.
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

from typing_extensions import deprecated

OTEL_COMPONENT_NAME: Final = "otel.component.name"
"""
A name uniquely identifying the instance of the OpenTelemetry component within its containing SDK instance.
Note: Implementations SHOULD ensure a low cardinality for this attribute, even across application or SDK restarts.
E.g. implementations MUST NOT use UUIDs as values for this attribute.

Implementations MAY achieve these goals by following a `<otel.component.type>/<instance-counter>` pattern, e.g. `batching_span_processor/0`.
Hereby `otel.component.type` refers to the corresponding attribute value of the component.

The value of `instance-counter` MAY be automatically assigned by the component and uniqueness within the enclosing SDK instance MUST be guaranteed.
For example, `<instance-counter>` MAY be implemented by using a monotonically increasing counter (starting with `0`), which is incremented every time an
instance of the given component type is started.

With this implementation, for example the first Batching Span Processor would have `batching_span_processor/0`
as `otel.component.name`, the second one `batching_span_processor/1` and so on.
These values will therefore be reused in the case of an application restart.
"""

OTEL_COMPONENT_TYPE: Final = "otel.component.type"
"""
A name identifying the type of the OpenTelemetry component.
Note: If none of the standardized values apply, implementations SHOULD use the language-defined name of the type.
E.g. for Java the fully qualified classname SHOULD be used in this case.
"""

OTEL_EVENT_NAME: Final = "otel.event.name"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OTEL_EVENT_NAME`.
"""

OTEL_LIBRARY_NAME: Final = "otel.library.name"
"""
Deprecated: Replaced by `otel.scope.name`.
"""

OTEL_LIBRARY_VERSION: Final = "otel.library.version"
"""
Deprecated: Replaced by `otel.scope.version`.
"""

OTEL_SCOPE_NAME: Final = "otel.scope.name"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OTEL_SCOPE_NAME`.
"""

OTEL_SCOPE_SCHEMA_URL: Final = "otel.scope.schema_url"
"""
The schema URL of the instrumentation scope.
"""

OTEL_SCOPE_VERSION: Final = "otel.scope.version"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OTEL_SCOPE_VERSION`.
"""

OTEL_SPAN_PARENT_ORIGIN: Final = "otel.span.parent.origin"
"""
Determines whether the span has a parent span, and if so, [whether it is a remote parent](https://opentelemetry.io/docs/specs/otel/trace/api/#isremote).
"""

OTEL_SPAN_SAMPLING_RESULT: Final = "otel.span.sampling_result"
"""
The result value of the sampler for this span.
"""

OTEL_STATUS_CODE: Final = "otel.status_code"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OTEL_STATUS_CODE`.
"""

OTEL_STATUS_DESCRIPTION: Final = "otel.status_description"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OTEL_STATUS_DESCRIPTION`.
"""


class OtelComponentTypeValues(Enum):
    BATCHING_SPAN_PROCESSOR = "batching_span_processor"
    """The builtin SDK batching span processor."""
    SIMPLE_SPAN_PROCESSOR = "simple_span_processor"
    """The builtin SDK simple span processor."""
    BATCHING_LOG_PROCESSOR = "batching_log_processor"
    """The builtin SDK batching log record processor."""
    SIMPLE_LOG_PROCESSOR = "simple_log_processor"
    """The builtin SDK simple log record processor."""
    OTLP_GRPC_SPAN_EXPORTER = "otlp_grpc_span_exporter"
    """OTLP span exporter over gRPC with protobuf serialization."""
    OTLP_HTTP_SPAN_EXPORTER = "otlp_http_span_exporter"
    """OTLP span exporter over HTTP with protobuf serialization."""
    OTLP_HTTP_JSON_SPAN_EXPORTER = "otlp_http_json_span_exporter"
    """OTLP span exporter over HTTP with JSON serialization."""
    ZIPKIN_HTTP_SPAN_EXPORTER = "zipkin_http_span_exporter"
    """Zipkin span exporter over HTTP."""
    OTLP_GRPC_LOG_EXPORTER = "otlp_grpc_log_exporter"
    """OTLP log record exporter over gRPC with protobuf serialization."""
    OTLP_HTTP_LOG_EXPORTER = "otlp_http_log_exporter"
    """OTLP log record exporter over HTTP with protobuf serialization."""
    OTLP_HTTP_JSON_LOG_EXPORTER = "otlp_http_json_log_exporter"
    """OTLP log record exporter over HTTP with JSON serialization."""
    PERIODIC_METRIC_READER = "periodic_metric_reader"
    """The builtin SDK periodically exporting metric reader."""
    OTLP_GRPC_METRIC_EXPORTER = "otlp_grpc_metric_exporter"
    """OTLP metric exporter over gRPC with protobuf serialization."""
    OTLP_HTTP_METRIC_EXPORTER = "otlp_http_metric_exporter"
    """OTLP metric exporter over HTTP with protobuf serialization."""
    OTLP_HTTP_JSON_METRIC_EXPORTER = "otlp_http_json_metric_exporter"
    """OTLP metric exporter over HTTP with JSON serialization."""
    PROMETHEUS_HTTP_TEXT_METRIC_EXPORTER = (
        "prometheus_http_text_metric_exporter"
    )
    """Prometheus metric exporter over HTTP with the default text-based format."""


class OtelSpanParentOriginValues(Enum):
    NONE = "none"
    """The span does not have a parent, it is a root span."""
    LOCAL = "local"
    """The span has a parent and the parent's span context [isRemote()](https://opentelemetry.io/docs/specs/otel/trace/api/#isremote) is false."""
    REMOTE = "remote"
    """The span has a parent and the parent's span context [isRemote()](https://opentelemetry.io/docs/specs/otel/trace/api/#isremote) is true."""


class OtelSpanSamplingResultValues(Enum):
    DROP = "DROP"
    """The span is not sampled and not recording."""
    RECORD_ONLY = "RECORD_ONLY"
    """The span is not sampled, but recording."""
    RECORD_AND_SAMPLE = "RECORD_AND_SAMPLE"
    """The span is sampled and recording."""


@deprecated(
    "Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OtelStatusCodeValues`."
)
class OtelStatusCodeValues(Enum):
    OK = "OK"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OtelStatusCodeValues.OK`."""
    ERROR = "ERROR"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.otel_attributes.OtelStatusCodeValues.ERROR`."""



import base64
import hashlib
import logging
from pathlib import Path

from cryptography.exceptions import InvalidSignature as CryptographyInvalidSig
from cryptography.hazmat.primitives.asymmetric import ec, padding, utils

from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.utils import x509
from salt.utils.hashutils import get_hash

log = logging.getLogger(__name__)

DEFAULT_HASHALG = "sha256"


class InvalidSignature(CommandExecutionError):
    """
    Raised when a signature is invalid.
    We save the file hash to avoid computing it multiple times for
    signed_by_any, if a path has been passed in and the signature algorithm
    supports prehashed data.
    """

    def __init__(self, *args, file_digest=None, data=None, pubkey=None):
        super().__init__(*args)
        self.file_digest = file_digest
        self.data = data
        self.pubkey = pubkey


def fingerprint(pubkey):
    """
    Return the SHA256 hexdigest of a pubkey's DER representation.

    pub
        The public key to calculate the fingerprint of.
        Can be any reference that can be passed to ``salt.utils.x509.load_pubkey``.
    """
    pubkey = x509.load_pubkey(pubkey)
    hsh = getattr(hashlib, DEFAULT_HASHALG)()
    hsh.update(x509.to_der(pubkey))
    return hsh.hexdigest()


def sign(privkey, data, digest=None, passphrase=None):
    """
    Sign data with a private key.

    privkey
        The private key to sign with. Can be any reference that can be passed
        to ``salt.utils.x509.load_privkey``.

    data
        The data to sign. Should be either ``str``, ``bytes`` or ``pathlib.Path`` object.

    digest
        The name of the hashing algorithm to use when creating signatures.
        Defaults to ``sha256``. Only relevant for ECDSA or RSA.

    passphrase
        If the private key is encrypted, the passphrase to decrypt it. Optional.
    """
    privkey = x509.load_privkey(privkey, passphrase=passphrase)
    key_type = x509.get_key_type(privkey)
    prehashed = False
    if isinstance(data, Path):
        if key_type in (x509.KEY_TYPE.RSA, x509.KEY_TYPE.EC):
            data = _get_file_digest(data, digest)
            prehashed = True
        else:
            data = data.read_bytes()
    elif isinstance(data, str):
        data = data.encode()
    if key_type == x509.KEY_TYPE.RSA:
        return _sign_rsa(privkey, data, digest, prehashed=prehashed)
    if key_type == x509.KEY_TYPE.EC:
        return _sign_ec(privkey, data, digest, prehashed=prehashed)
    if key_type == x509.KEY_TYPE.ED25519:
        return _sign_ed25519(privkey, data)
    if key_type == x509.KEY_TYPE.ED448:
        return _sign_ed448(privkey, data)
    raise CommandExecutionError(f"Unknown private key type: {privkey.__class__}")


def try_base64(data):
    """
    Check if the data is valid base64 and return the decoded
    bytes if so, otherwise return the data untouched.
    """
    try:
        if isinstance(data, str):
            data = data.encode("ascii", "strict")
        elif isinstance(data, bytes):
            pass
        else:
            raise CommandExecutionError("is_base64 only works with strings and bytes")
        decoded = base64.b64decode(data)
        if base64.b64encode(decoded) == data.replace(b"\n", b""):
            return decoded
        return data
    except (TypeError, ValueError):
        return data


def load_sig(sig):
    """
    Try to load an input that represents a signature into the signature's bytes.

    sig
        The reference to load. Can either be a base64-encoded string or a path
        to a local file in base64 encoding or raw bytes.
    """
    if x509.isfile(sig):
        sig = Path(sig).read_bytes()
    sig = try_base64(sig)
    if isinstance(sig, bytes):
        return sig
    raise CommandExecutionError(
        f"Failed loading signature '{sig}' as file and/or base64 string"
    )


def verify(pubkey, sig, data, digest=None, file_digest=None):
    """
    Verify a signature against a public key.

    On success, returns a tuple of (data, file_digest), which can reused
    by the callee when multiple signatures are checked against the same file.
    This avoids reading the file/calculating the digest multiple times.

    On failure, raises an InvalidSignature exception which carries
    ``data`` and ``file_digest`` attributes with the corresponding values.

    pub
        The public key to verify the signature against.
        Can be any reference that can be passed to ``salt.utils.x509.load_pubkey``.

    sig
        The signature to verify.
        Can be any reference that can be passed to ``salt.utils.asymmetric.load_sig``.

    data
        The data to sign. Should be either ``str``, ``bytes`` or ``pathlib.Path`` object.
        Ignored when ``file_digest`` is passed and the signing algorithm is either
        ECDSA or RSA.

    digest
        The name of the hashing algorithm to use when creating signatures.
        Defaults to ``sha256``. Only relevant for ECDSA or RSA.

    file_digest
        The ECDSA and RSA algorithms can be invoked with a precalculated digest
        in order to avoid loading the whole file into memory. This happens automatically
        during the execution of this function, but when checking multiple signatures,
        you can cache the calculated value and pass it back in.
    """
    pubkey = x509.load_pubkey(pubkey)
    signature = load_sig(sig)
    key_type = x509.get_key_type(pubkey)
    sig_data = data
    if key_type in (x509.KEY_TYPE.RSA, x509.KEY_TYPE.EC):
        if file_digest:
            sig_data = file_digest
        elif isinstance(data, Path):
            file_digest = sig_data = _get_file_digest(data, digest)
        elif isinstance(data, str):
            sig_data = data.encode()
    elif isinstance(data, Path):
        data = sig_data = data.read_bytes()
    elif isinstance(data, str):
        data = sig_data = data.encode()

    try:
        if key_type == x509.KEY_TYPE.RSA:
            _verify_rsa(
                pubkey,
                signature,
                sig_data,
                digest or DEFAULT_HASHALG,
                prehashed=bool(file_digest),
            )
            return data, file_digest
        if key_type == x509.KEY_TYPE.EC:
            _verify_ec(
                pubkey,
                signature,
                sig_data,
                digest or DEFAULT_HASHALG,
                prehashed=bool(file_digest),
            )
            return data, file_digest
        if key_type == x509.KEY_TYPE.ED25519:
            _verify_ed25519(pubkey, signature, sig_data)
            return data, file_digest
        if key_type == x509.KEY_TYPE.ED448:
            _verify_ed448(pubkey, signature, sig_data)
            return data, file_digest
    except CryptographyInvalidSig as err:
        raise InvalidSignature(
            f"Invalid signature for key {fingerprint(pubkey)}",
            file_digest=file_digest,
            data=data,
            pubkey=pubkey,
        ) from err
    raise SaltInvocationError(f"Unknown public key type: {pubkey.__class__}")


def _sign_rsa(priv, data, digest, prehashed=False):
    pad_hashalg = sig_hashalg = x509.get_hashing_algorithm(digest or DEFAULT_HASHALG)
    if prehashed:
        sig_hashalg = utils.Prehashed(sig_hashalg)
    return priv.sign(
        data,
        padding.PSS(mgf=padding.MGF1(pad_hashalg), salt_length=padding.PSS.MAX_LENGTH),
        sig_hashalg,
    )


def _sign_ec(priv, data, digest, prehashed=False):
    hashalg = x509.get_hashing_algorithm(digest or DEFAULT_HASHALG)
    if prehashed:
        hashalg = utils.Prehashed(hashalg)
    return priv.sign(data, ec.ECDSA(hashalg))


def _sign_ed25519(priv, data):
    return priv.sign(data)


def _sign_ed448(priv, data):
    return priv.sign(data)


def _verify_rsa(pub, sig, data, digest=DEFAULT_HASHALG, prehashed=False):
    pad_hashalg = sig_hashalg = x509.get_hashing_algorithm(digest)
    if prehashed:
        sig_hashalg = utils.Prehashed(sig_hashalg)
    # Technically, scheme hash function and the MGF hash function can be different,
    # but that's not common practice.
    pad = padding.PSS(mgf=padding.MGF1(pad_hashalg), salt_length=padding.PSS.AUTO)
    pub.verify(sig, data, pad, sig_hashalg)


def _verify_ec(pub, sig, data, digest=DEFAULT_HASHALG, prehashed=False):
    hashalg = x509.get_hashing_algorithm(digest)
    if prehashed:
        hashalg = utils.Prehashed(hashalg)
    pub.verify(sig, data, ec.ECDSA(hashalg))


def _verify_ed25519(pub, sig, data):
    pub.verify(sig, data)


def _verify_ed448(pub, sig, data):
    pub.verify(sig, data)


def _get_file_digest(file, digest):
    hexdigest = get_hash(file, digest or DEFAULT_HASHALG)
    return bytes.fromhex(hexdigest)



"""
The classes and functions in this module are related to requisite
handling and ordering of chunks by the State compiler
"""

from __future__ import annotations

import fnmatch
import logging
import sys
from collections import defaultdict
from collections.abc import Generator, Iterable, Sequence
from enum import Enum, auto
from heapq import heappop, heappush
from typing import TYPE_CHECKING, Any

log = logging.getLogger(__name__)

# See https://docs.saltproject.io/en/latest/ref/states/layers.html for details on the naming
LowChunk = dict[str, Any]

if TYPE_CHECKING or sys.version_info >= (3, 10):
    staticmethod_hack = staticmethod
else:
    # Python < 3.10 does not support calling static methods directly from the class body
    # as is the case with enum _generate_next_value_.
    # Since @staticmethod is only added for static type checking, substitute a dummy decorator.
    def staticmethod_hack(f):
        return f


def _gen_tag(low: LowChunk) -> str:
    """
    Generate the a unique identifer tag string from the low data structure
    """
    return "{0[state]}_|-{0[__id__]}_|-{0[name]}_|-{0[fun]}".format(low)


def trim_req(req: dict[str, Any]) -> dict[str, Any]:
    """
    Trim any function off of a requisite reference
    """
    reqfirst, valfirst = next(iter(req.items()))
    if "." in reqfirst:
        return {reqfirst.split(".", maxsplit=1)[0]: valfirst}
    return req


class RequisiteType(str, Enum):
    """Types of direct requisites"""

    # Once salt no longer needs to support python < 3.10,
    # remove this hack and use @staticmethod
    @staticmethod_hack
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> tuple[str, int]:
        return name.lower(), count

    def __new__(cls, value, weight):
        member = str.__new__(cls, value)
        member._value_ = value
        member.weight = weight
        return member

    def __init__(self, value, weight):
        super().__init__()
        self._value_ = value
        self.weight = weight

    def __str__(self):
        return self.value

    # The items here are listed in order of precedence for determining
    # the order of execution, so do not change the order unless you
    # are intentionally changing the precedence
    ONFAIL = auto()
    ONFAIL_ANY = auto()
    ONFAIL_ALL = auto()
    REQUIRE = auto()
    REQUIRE_ANY = auto()
    ONCHANGES = auto()
    ONCHANGES_ANY = auto()
    WATCH = auto()
    WATCH_ANY = auto()
    PREREQ = auto()
    PREREQUIRED = auto()
    LISTEN = auto()


class DiGraphCycle(Exception):
    """
    Custom DiGraphCycle exception raised on detecting cycle.
    """


class MultiDiGraph:
    """
    Custom multigraph implementation replacing networkx.MultiDiGraph.

    A directed multigraph allows multiple edges between the same pair of nodes,
    each identified by a unique key. This implementation provides the specific
    API needed by the DependencyGraph class.
    """

    def __init__(self):
        # Node attributes: node_id -> {attr_name: value}
        self._nodes: dict[str, dict[str, Any]] = {}

        # Adjacency lists for efficient edge traversal
        # source -> target -> edge_key -> edge_data
        self._out_edges: dict[str, dict[str, dict[Any, dict[str, Any]]]] = defaultdict(
            lambda: defaultdict(dict)
        )
        self._in_edges: dict[str, dict[str, dict[Any, dict[str, Any]]]] = defaultdict(
            lambda: defaultdict(dict)
        )

    def __bool__(self):
        """
        Return true if MultiDiGraph is not empty.
        """
        return bool(self._nodes)

    @property
    def nodes(self) -> dict[str, dict[str, Any]]:
        """
        Return the nodes dictionary for attribute access.
        """
        return self._nodes

    def add_node(self, node_id: str, **attrs) -> None:
        """
        Add a node with optional attributes.
        """
        if node_id not in self._nodes:
            self._nodes[node_id] = {}
        self._nodes[node_id].update(attrs)

    def add_edge(self, source: str, target: str, key: Any, **data) -> None:
        """
        Add an edge from source to target with a given key and optional data.

        Args:
            source: Source node identifier
            target: Target node identifier
            key: Edge key (allows multiple edges between same node pair)
            **data: Optional edge attributes
        """
        # Ensure nodes exist
        if source not in self._nodes:
            self.add_node(source)
        if target not in self._nodes:
            self.add_node(target)

        # Add edge to both adjacency lists
        self._out_edges[source][target][key] = data
        self._in_edges[target][source][key] = data

    def in_edges(
        self, node: str, keys: bool = False, data: bool = False
    ) -> Generator[tuple]:
        """
        Return an iterator over the incoming edges of node.

        Args:
            node: The node identifier
            keys: If True, include edge keys in the output
            data: If True, include edge data in the output

        Yields:
            Tuples of (source, target) if keys=False and data=False
            Tuples of (source, target, key) if keys=True and data=False
            Tuples of (source, target, key, edge_data) if keys=True and data=True
        """
        for source, keys_dict in self._in_edges.get(node, {}).items():
            for edge_key, edge_data in keys_dict.items():
                if keys and data:
                    yield (source, node, edge_key, edge_data)
                elif keys:
                    yield (source, node, edge_key)
                else:
                    yield (source, node)

    def out_edges(
        self, node: str, keys: bool = False, data: bool = False
    ) -> Generator[tuple]:
        """
        Return an iterator over the outgoing edges of node.

        Args:
            node: The node identifier
            keys: If True, include edge keys in the output
            data: If True, include edge data in the output

        Yields:
            Tuples of (source, target) if keys=False and data=False
            Tuples of (source, target, key) if keys=True and data=False
            Tuples of (source, target, key, edge_data) if keys=True and data=True
        """
        for target, keys_dict in self._out_edges.get(node, {}).items():
            for edge_key, edge_data in keys_dict.items():
                if keys and data:
                    yield (node, target, edge_key, edge_data)
                elif keys:
                    yield (node, target, edge_key)
                else:
                    yield (node, target)

    def has_path(self, source: str, target: str) -> bool:
        """
        Check if there is a path from source to target using DFS.

        Args:
            source: Source node identifier
            target: Target node identifier

        Returns:
            True if a path exists, False otherwise
        """
        if source == target:
            return True
        if source not in self._nodes or target not in self._nodes:
            return False

        visited = set()
        stack = [source]

        while stack:
            node = stack.pop()
            if node == target:
                return True
            if node in visited:
                continue
            visited.add(node)

            for child in self._out_edges.get(node, {}):
                if child not in visited:
                    stack.append(child)

        return False

    def find_cycle(self) -> list[tuple[str, str, Any]]:
        """
        Find a cycle in the graph using DFS with back-edge detection.

        Returns:
            List of (source, target, key) tuples forming the cycle.
            Empty list if no cycle exists.
        """
        WHITE, GRAY, BLACK = 0, 1, 2
        color = {node: WHITE for node in self._nodes}
        parent = {}
        parent_key = {}

        def dfs_visit(node):
            color[node] = GRAY
            for target, keys_dict in self._out_edges.get(node, {}).items():
                for edge_key in keys_dict:
                    if color[target] == GRAY:
                        # Found back edge - reconstruct cycle
                        cycle = [(node, target, edge_key)]
                        current = node
                        while current != target:
                            prev = parent[current]
                            cycle.append((prev, current, parent_key[current]))
                            current = prev
                        return list(reversed(cycle))
                    elif color[target] == WHITE:
                        parent[target] = node
                        parent_key[target] = edge_key
                        if cycle := dfs_visit(target):
                            return cycle
            color[node] = BLACK
            return None

        for node in self._nodes:
            if color[node] == WHITE:
                if cycle := dfs_visit(node):
                    return cycle

        return []

    def lexicographical_topological_sort(self, key):
        """
        Topological sort with lexicographical ordering via custom key function.

        Uses Kahn's algorithm with a priority queue to ensure nodes are processed
        in lexicographical order when multiple nodes have no remaining dependencies.

        Args:
            key: A function that takes a node identifier and returns a sort key

        Yields:
            Node identifiers in topologically sorted, lexicographical order

        Raises:
            ValueError: If the graph contains a cycle
        """
        # Capture the initial node count before iteration
        # (the graph may be modified during iteration)
        initial_node_count = len(self._nodes)

        # Calculate in-degrees (count edges, not just neighbors, for multigraph)
        in_degree = {node: 0 for node in self._nodes}
        for target, source_edges in self._in_edges.items():
            for source, keys_dict in source_edges.items():
                in_degree[target] += len(keys_dict)

        # Priority queue: (key_value, node)
        heap = []
        for node in self._nodes:
            if in_degree[node] == 0:
                heappush(heap, (key(node), node))

        processed_count = 0
        while heap:
            _, node = heappop(heap)
            yield node
            processed_count += 1

            # Reduce in-degree for neighbors
            for target, keys_dict in self._out_edges.get(node, {}).items():
                in_degree[target] -= len(keys_dict)
                if in_degree[target] == 0:
                    heappush(heap, (key(target), target))

        if processed_count != initial_node_count:
            # Cycle detected
            raise DiGraphCycle("Graph contains a cycle")


class DependencyGraph:
    """
    Class used to track dependencies (requisites) among salt states.

    This class utilizes a Directed Acyclic Graph to determine the
    ordering of states. The nodes represent the individual states that
    can be depended on and edges represent the types of requisites
    between the states.
    """

    __slots__ = ("dag", "nodes_lookup_map", "sls_to_nodes")

    def __init__(self) -> None:
        self.dag = MultiDiGraph()
        # a mapping to node_id to be able to find nodes with
        # specific state type (module name), names, and/or IDs
        self.nodes_lookup_map: dict[tuple[str, str], set[str]] = {}
        self.sls_to_nodes: dict[str, set[str]] = {}

    def _add_prereq(self, node_tag: str, req_tag: str):
        # the prerequiring chunk is the state declaring the prereq
        # requisite; the prereq/prerequired state is the one that is
        # declared in the requisite prereq statement
        self.dag.nodes[node_tag]["chunk"]["__prerequiring__"] = True
        prereq_chunk = self.dag.nodes[req_tag]["chunk"]
        # set __prereq__ true to run the state in test mode
        prereq_chunk["__prereq__"] = True
        prereq_check_node = self._get_prereq_node_tag(req_tag)
        if not self.dag.nodes.get(prereq_check_node):
            self.dag.add_node(
                prereq_check_node, chunk=prereq_chunk, state=prereq_chunk["state"]
            )
            # all the dependencies of the node for the prerequired
            # chunk also need to be applied to its prereq check node
            for dependency_node, _, req_type, data in self.dag.in_edges(
                req_tag, data=True, keys=True
            ):
                if req_type != RequisiteType.PREREQ:
                    self.dag.add_edge(
                        dependency_node, prereq_check_node, req_type, **data
                    )
        self.dag.add_edge(prereq_check_node, node_tag, RequisiteType.PREREQ)
        self.dag.add_edge(node_tag, req_tag, RequisiteType.REQUIRE)

    def _add_reqs(
        self,
        node_tag: str,
        has_preq_node: bool,
        req_type: RequisiteType,
        req_tags: Iterable[str],
    ) -> None:
        for req_tag in req_tags:
            if req_type == RequisiteType.PREREQ:
                self._add_prereq(node_tag, req_tag)
            else:
                if has_preq_node:
                    # if the low chunk is set to run in test mode for a
                    # prereq check then also add the requisites to the
                    # prereq node.
                    prereq_node_tag = self._get_prereq_node_tag(node_tag)
                    self.dag.add_edge(req_tag, prereq_node_tag, key=req_type)
                self.dag.add_edge(req_tag, node_tag, key=req_type)

    def _copy_edges(self, source: str, dest: str) -> None:
        """Add the edges from source node to dest node"""
        for dependency, _, req_type, data in self.dag.in_edges(
            source, data=True, keys=True
        ):
            self.dag.add_edge(dependency, dest, req_type, **data)
        for _, dependent, req_type, data in self.dag.out_edges(
            source, data=True, keys=True
        ):
            self.dag.add_edge(dest, dependent, req_type, **data)

    def _get_chunk_order(self, cap: int, node: str) -> tuple[int | float, int | float]:
        dag = self.dag
        stack: list[tuple[str, bool, int | float, int | float]] = [
            # node, is_processing_children, child_min, req_order
            (node, False, float("inf"), float("-inf"))
        ]
        order = cap
        while stack:
            node, is_processing_children, child_min, req_order = stack[-1]
            node_data = dag.nodes[node]
            chunk = node_data.get("chunk", {})
            if not is_processing_children:  # initial stage
                order = chunk.get("order")
                if order is None or not isinstance(order, (int, float)):
                    if order == "last":
                        order = cap + 1000000
                    elif order == "first":
                        order = 0
                    else:
                        order = cap
                    chunk["order"] = order
                name_order = chunk.pop("name_order", 0)
                if name_order:
                    order += name_order / 10000.0
                    chunk["order"] = order
                if order < 0:
                    order += cap + 1000000
                    chunk["order"] = order
                stack.pop()
                # update stage
                stack.append((node, True, child_min, req_order))
            else:  # after processing node
                child_min_node = node_data.get("child_min")
                if child_min_node is None:
                    for _, child, req_type in dag.out_edges(node, keys=True):
                        if req_order <= req_type.weight:
                            req_order = req_type.weight
                            child_order = (
                                dag.nodes[child]
                                .get("chunk", {})
                                .get("order", float("inf"))
                            )
                            if child_order is None or not isinstance(
                                child_order, (int, float)
                            ):
                                if child_order == "last":
                                    child_order = cap + 1000000
                                elif child_order == "first":
                                    child_order = 0
                                else:
                                    child_order = cap
                                dag.nodes[child]["chunk"]["order"] = child_order
                            child_min = min(child_min, child_order)
                    node_data["child_min"] = child_min
                    if order > child_min:
                        order = child_min
                stack.pop()
        return (order, chunk["order"])

    def _get_prereq_node_tag(self, low_tag: str):
        return f"{low_tag}_|-__prereq_test__"

    def _is_fnmatch_pattern(self, value: str) -> bool:
        return any(char in value for char in ("*", "?", "[", "]"))

    def _chunk_str(self, chunk: LowChunk) -> str:
        node_dict = {
            "SLS": chunk["__sls__"],
            "ID": chunk["__id__"],
        }
        if chunk["__id__"] != chunk["name"]:
            node_dict["NAME"] = chunk["name"]
        return str(node_dict)

    def add_chunk(self, low: LowChunk, allow_aggregate: bool) -> None:
        node_id = _gen_tag(low)
        self.dag.add_node(
            node_id, allow_aggregate=allow_aggregate, chunk=low, state=low["state"]
        )
        self.nodes_lookup_map.setdefault((low["state"], low["name"]), set()).add(
            node_id
        )
        self.nodes_lookup_map.setdefault((low["state"], low["__id__"]), set()).add(
            node_id
        )
        self.nodes_lookup_map.setdefault(("id", low["__id__"]), set()).add(node_id)
        self.nodes_lookup_map.setdefault(("id", low["name"]), set()).add(node_id)
        if sls := low.get("__sls__"):
            self.sls_to_nodes.setdefault(sls, set()).add(node_id)
        if sls_included_from := low.get("__sls_included_from__"):
            for sls in sls_included_from:
                self.sls_to_nodes.setdefault(sls, set()).add(node_id)

    def add_dependency(
        self, low: LowChunk, req_type: RequisiteType, req_key: str, req_val: str
    ) -> bool:
        found = False
        has_prereq_node = low.get("__prereq__", False)
        if req_key == "sls":
            # Allow requisite tracking of entire sls files
            if self._is_fnmatch_pattern(req_val):
                found = True
                node_tag = _gen_tag(low)
                for sls, req_tags in self.sls_to_nodes.items():
                    if fnmatch.fnmatch(sls, req_val):
                        found = True
                        self._add_reqs(node_tag, has_prereq_node, req_type, req_tags)
            else:
                node_tag = _gen_tag(low)
                if req_tags := self.sls_to_nodes.get(req_val, []):
                    found = True
                    self._add_reqs(node_tag, has_prereq_node, req_type, req_tags)
        elif self._is_fnmatch_pattern(req_val):
            # This iterates over every chunk to check
            # if any match instead of doing a look up since
            # it has to support wildcard matching.
            node_tag = _gen_tag(low)
            for (state_type, name_or_id), req_tags in self.nodes_lookup_map.items():
                if req_key == state_type and (fnmatch.fnmatch(name_or_id, req_val)):
                    found = True
                    self._add_reqs(node_tag, has_prereq_node, req_type, req_tags)
        elif req_tags := self.nodes_lookup_map.get((req_key, req_val)):
            found = True
            node_tag = _gen_tag(low)
            self._add_reqs(node_tag, has_prereq_node, req_type, req_tags)
        return found

    def add_requisites(self, low: LowChunk, disabled_reqs: Sequence[str]) -> str | None:
        """
        Add all the dependency requisites of the low chunk as edges to the DAG
        :return: an error string if there was an error otherwise None
        """
        present = False
        for req_type in RequisiteType:
            if req_type.value in low:
                present = True
                break
        if not present:
            return None
        reqs = {
            rtype: []
            for rtype in (
                RequisiteType.REQUIRE,
                RequisiteType.REQUIRE_ANY,
                RequisiteType.WATCH,
                RequisiteType.WATCH_ANY,
                RequisiteType.PREREQ,
                RequisiteType.ONFAIL,
                RequisiteType.ONFAIL_ANY,
                RequisiteType.ONFAIL_ALL,
                RequisiteType.ONCHANGES,
                RequisiteType.ONCHANGES_ANY,
            )
        }
        for r_type in reqs:
            if low_reqs := low.get(r_type.value):
                if r_type in disabled_reqs:
                    log.warning("The %s requisite has been disabled, Ignoring.", r_type)
                    continue
                for req_ref in low_reqs:
                    if isinstance(req_ref, str):
                        req_ref = {"id": req_ref}
                    req_ref = trim_req(req_ref)
                    # req_key: match state module name
                    # req_val: match state id or name
                    req_key, req_val = next(iter(req_ref.items()))
                    if req_val is None:
                        continue
                    if not isinstance(req_val, str):
                        return (
                            f"Requisite [{r_type}: {req_key}] in state"
                            f" [{low['name']}] in SLS [{low.get('__sls__')}]"
                            " must have a string as the value"
                        )
                    found = self.add_dependency(low, r_type, req_key, req_val)
                    if not found:
                        return (
                            "Referenced state does not exist"
                            f" for requisite [{r_type}: ({req_key}: {req_val})] in state"
                            f" [{low['name']}] in SLS [{low.get('__sls__')}]"
                        )
        return None

    def aggregate_and_order_chunks(self, cap: int) -> list[LowChunk]:
        """
        Aggregate eligible nodes in the dependencies graph.

        Return a list of the chunks in the sorted order in which the
        chunks should be executed.
        Nodes are eligible for aggregation if the state function in the
        chunks match and aggregation is enabled in the configuration for
        the state function.
        :param cap: the maximum order value configured in the states
        :return: the ordered chunks
        """
        dag: MultiDiGraph = self.dag
        # dict for tracking topo order and for mapping each node that
        # was aggregated to the aggregated node that replaces it
        topo_order = {}

        max_group_size = 500
        groups_by_type = defaultdict(list)

        def _get_order(node):
            chunk = dag.nodes[node].get("chunk", {})
            chunk_label = "{0[state]}{0[name]}{0[fun]}".format(chunk) if chunk else ""
            chunk_order = self._get_chunk_order(cap, node)
            return (chunk_order, chunk_label)

        # Iterate over the nodes in topological order to get the correct
        # ordering which takes requisites into account
        for node in dag.lexicographical_topological_sort(key=_get_order):
            topo_order[node] = None
            data = dag.nodes[node]
            if not data.get("allow_aggregate"):
                continue

            node_type = data["state"]
            added = False
            for idx, group in enumerate(groups_by_type[node_type]):
                if len(group) >= max_group_size:
                    continue
                # Check if the node can be reached from any node in the group
                first_node = next(iter(group))
                agg_node = topo_order.get(first_node)
                # Since we are iterating in topological order we know
                # that there is no path from the current node to the
                # node in the group; so we only need to check the path
                # from the group node to the current node
                reachable = dag.has_path(agg_node or first_node, node)
                if not reachable:
                    # If not, add the node to the group
                    if agg_node is None:
                        # there is now more than one node for this
                        # group so aggregate them
                        agg_node = f"__aggregate_{node_type}_{idx}__"
                        dag.add_node(
                            agg_node, state=node_type, aggregated_nodes=group.keys()
                        )
                        # add the edges of the first node in the group to
                        # the aggregate
                        self._copy_edges(first_node, agg_node)
                        dag.nodes[first_node]["aggregate"] = agg_node
                        topo_order[first_node] = agg_node

                    self._copy_edges(node, agg_node)
                    dag.nodes[node]["aggregate"] = agg_node
                    topo_order[node] = agg_node
                    group[node] = None
                    added = True
                    break

            # If the node was not added to any set, create a new set
            if not added:
                # use a dict instead of set to retain insertion ordering
                groups_by_type[node_type].append({node: None})

        ordered_chunks = [dag.nodes[node].get("chunk", {}) for node in topo_order]
        return ordered_chunks

    def find_cycle_edges(self) -> list[tuple[LowChunk, RequisiteType, LowChunk]]:
        """
        Find the cycles if the graph is not a Directed Acyclic Graph
        """
        cycle_edges = []
        for dependency, dependent, req_type in self.dag.find_cycle():
            dependency_chunk = self.dag.nodes[dependency]["chunk"]
            dependent_chunk = self.dag.nodes[dependent]["chunk"]
            if req_type not in dependent_chunk and req_type == RequisiteType.REQUIRE:
                # show the original prereq requisite for the require edges
                # added for the prereq
                req_type = RequisiteType.PREREQ
            cycle_edges.append((dependent_chunk, req_type, dependency_chunk))
        return cycle_edges

    def get_aggregate_chunks(self, low: LowChunk) -> list[LowChunk]:
        """
        Get the chunks that were set to be valid for aggregation with
        this low chunk.
        """
        low_tag = _gen_tag(low)
        if aggregate_node := self.dag.nodes[low_tag].get("aggregate"):
            return [
                self.dag.nodes[node]["chunk"]
                for node in self.dag.nodes[aggregate_node]["aggregated_nodes"]
            ]
        return []

    def get_cycles_str(self) -> str:
        cycle_edges = [
            f"({self._chunk_str(dependency)}, '{req_type.value}', {self._chunk_str(dependent)})"
            for dependency, req_type, dependent in self.find_cycle_edges()
        ]
        return ", ".join(cycle_edges)

    def to_dot(self) -> str:
        """
        Return a DOT representation of the dependency graph
        """
        dot = ["digraph G {"]
        # Add nodes
        for node_id, node_data in self.dag.nodes.items():
            chunk = node_data.get("chunk")
            if chunk:
                label = f"{chunk['state']}.{chunk['fun']}\\n{chunk['__id__']}"
                if chunk["name"] != chunk["__id__"]:
                    label += f"\\n({chunk['name']})"
                dot.append(f'  "{node_id}" [label="{label}"];')
            elif "aggregated_nodes" in node_data:
                dot.append(
                    f'  "{node_id}" [label="Aggregate {node_data["state"]}", shape=box];'
                )
            else:
                dot.append(f'  "{node_id}" [label="{node_id}"];')

        # Add edges
        for source_node in self.dag.nodes:
            for source, target, req_type in self.dag.out_edges(source_node, keys=True):
                dot.append(f'  "{source}" -> "{target}" [label="{req_type}"];')

        dot.append("}")
        return "\n".join(dot)

    def get_dependencies(
        self, low: LowChunk
    ) -> Generator[tuple[RequisiteType, LowChunk]]:
        """Get the requisite type and low chunk for each dependency of low"""
        low_tag = _gen_tag(low)
        if low.get("__prereq__"):
            # if the low chunk is set to run in test mode for a
            # prereq check then return the reqs for prereq test node.
            low_tag = self._get_prereq_node_tag(low_tag)
        for req_id, _, req_type in self.dag.in_edges(low_tag, keys=True):
            if chunk := self.dag.nodes[req_id].get("chunk"):
                yield req_type, chunk
            else:
                for node in self.dag.nodes[req_id]["aggregated_nodes"]:
                    yield req_type, self.dag.nodes[node].get("chunk")



"""
Event constants used by listeners and servers, to be imported elsewhere in Salt code.

Do NOT, import any salt modules (salt.utils, salt.config, etc.) into this file,
as this may result in circular imports.

.. versionadded:: 3000
"""

# Constants for events on the minion bus
MINION_PILLAR_REFRESH_COMPLETE = "/salt/minion/minion_pillar_refresh_complete"
MINION_MOD_REFRESH_COMPLETE = "/salt/minion/minion_mod_refresh_complete"
